from util import *
from vonage.errors import MeetingsError, ClientError, ServerError

import responses
import json
from pytest import raises


@responses.activate
def test_create_instant_room(meetings, dummy_data):
    stub(
        responses.POST,
        "https://api-eu.vonage.com/beta/meetings/rooms",
        fixture_path='meetings/meeting_room.json',
    )

    params = {'display_name': 'my_test_room'}
    meeting = meetings.create_room(params)

    assert isinstance(meeting, dict)
    assert request_user_agent() == dummy_data.user_agent
    assert meeting['id'] == 'b3142c46-d1c1-4405-baa6-85683827ed69'
    assert meeting['display_name'] == 'my_test_room'
    assert meeting['expires_at'] == '2023-01-24T03:30:38.629Z'
    assert meeting['join_approval_level'] == 'none'


def test_create_instant_room_error_expiry(meetings, dummy_data):
    params = {'display_name': 'my_test_room', 'expires_at': '2023-01-24T03:30:38.629Z'}
    with raises(MeetingsError) as err:
        meetings.create_room(params)
    assert str(err.value) == 'Cannot set "expires_at" for an instant room.'


@responses.activate
def test_create_long_term_room(meetings, dummy_data):
    stub(
        responses.POST,
        "https://api-eu.vonage.com/beta/meetings/rooms",
        fixture_path='meetings/long_term_room.json',
    )

    params = {
        'display_name': 'test_long_term_room',
        'type': 'long_term',
        'expires_at': '2023-01-30T00:47:04+0000',
    }
    meeting = meetings.create_room(params)

    assert isinstance(meeting, dict)
    assert request_user_agent() == dummy_data.user_agent
    assert meeting['id'] == '33791484-231c-421b-8349-96e1a44e27d2'
    assert meeting['display_name'] == 'test_long_term_room'
    assert meeting['expires_at'] == '2023-01-30T00:47:04.000Z'


def test_create_room_error(meetings):
    with raises(MeetingsError) as err:
        meetings.create_room()
        assert (
            str(err.value)
            == 'You must include a value for display_name as a field in the params dict when creating a meeting room.'
        )


def test_create_long_term_room_error(meetings):
    params = {
        'display_name': 'test_long_term_room',
        'type': 'long_term',
    }
    with raises(MeetingsError) as err:
        meetings.create_room(params)
    assert str(err.value) == 'You must set a value for "expires_at" for a long-term room.'


@responses.activate
def test_get_room(meetings):
    stub(
        responses.GET,
        'https://api-eu.vonage.com/beta/meetings/rooms/b3142c46-d1c1-4405-baa6-85683827ed69',
        fixture_path='meetings/meeting_room.json',
    )
    meeting = meetings.get_room(room_id='b3142c46-d1c1-4405-baa6-85683827ed69')

    assert isinstance(meeting, dict)
    assert meeting['id'] == 'b3142c46-d1c1-4405-baa6-85683827ed69'
    assert meeting['display_name'] == 'my_test_room'
    assert meeting['expires_at'] == '2023-01-24T03:30:38.629Z'
    assert meeting['join_approval_level'] == 'none'
    assert meeting['ui_settings']['language'] == 'default'
    assert meeting['available_features']['is_locale_switcher_available'] == False
    assert meeting['available_features']['is_captions_available'] == False


def test_get_room_error_no_room_specified(meetings):
    with raises(TypeError):
        meetings.get_room()


@responses.activate
def test_list_rooms(meetings):
    stub(
        responses.GET,
        'https://api-eu.vonage.com/beta/meetings/rooms',
        fixture_path='meetings/multiple_rooms.json',
    )
    response = meetings.list_rooms()

    assert isinstance(response, dict)
    assert response['_embedded'][0]['id'] == '4814804d-7c2d-4846-8c7d-4f6fae1f910a'
    assert response['_embedded'][1]['id'] == 'de34416a-2a4c-4a59-a16a-8cd7d3121ea0'
    assert response['_embedded'][2]['id'] == 'd44529db-d1fa-48d5-bba0-43034bf91ae4'
    assert response['_embedded'][3]['id'] == '4f7dc750-6049-42ef-a25f-e7afa4953e32'
    assert response['_embedded'][4]['id'] == 'b3142c46-d1c1-4405-baa6-85683827ed69'
    assert response['total_items'] == 5


@responses.activate
def test_list_rooms_with_page_size(meetings):
    stub(
        responses.GET,
        'https://api-eu.vonage.com/beta/meetings/rooms',
        fixture_path='meetings/multiple_fewer_rooms.json',
    )
    response = meetings.list_rooms(page_size=2)

    assert isinstance(response, dict)
    assert response['_embedded'][0]['id'] == '4814804d-7c2d-4846-8c7d-4f6fae1f910a'
    assert response['_embedded'][1]['id'] == 'de34416a-2a4c-4a59-a16a-8cd7d3121ea0'
    assert response['page_size'] == 2
    assert response['total_items'] == 2


@responses.activate
def test_error_unauthorized(meetings):
    stub(
        responses.GET,
        'https://api-eu.vonage.com/beta/meetings/rooms',
        fixture_path='meetings/unauthorized.json',
        status_code=401,
    )
    with raises(ClientError) as err:
        meetings.list_rooms()
    assert str(err.value) == 'Authentication failed.'


@responses.activate
def test_update_room(meetings):
    stub(
        responses.PATCH,
        'https://api-eu.vonage.com/beta/meetings/rooms/b3142c46-d1c1-4405-baa6-85683827ed69',
        fixture_path='meetings/update_room.json',
    )

    params = {
        'update_details': {
            "available_features": {
                "is_recording_available": False,
                "is_chat_available": False,
                "is_whiteboard_available": False,
            }
        }
    }
    meeting = meetings.update_room(room_id='b3142c46-d1c1-4405-baa6-85683827ed69', params=params)

    assert meeting['id'] == '33791484-231c-421b-8349-96e1a44e27d2'
    assert meeting['available_features']['is_recording_available'] == False
    assert meeting['available_features']['is_chat_available'] == False
    assert meeting['available_features']['is_whiteboard_available'] == False


@responses.activate
def test_add_theme_to_room(meetings):
    stub(
        responses.PATCH,
        'https://api-eu.vonage.com/beta/meetings/rooms/33791484-231c-421b-8349-96e1a44e27d2',
        fixture_path='meetings/long_term_room_with_theme.json',
    )

    meeting = meetings.add_theme_to_room(
        room_id='33791484-231c-421b-8349-96e1a44e27d2',
        theme_id='90a21428-b74a-4221-adc3-783935d654db',
    )

    assert meeting['id'] == '33791484-231c-421b-8349-96e1a44e27d2'
    assert meeting['theme_id'] == '90a21428-b74a-4221-adc3-783935d654db'


@responses.activate
def test_update_room_error_no_room_specified(meetings):
    stub(
        responses.PATCH,
        'https://api-eu.vonage.com/beta/meetings/rooms/b3142c46-d1c1-4405-baa6-85683827ed69',
        fixture_path='meetings/update_room_type_error.json',
        status_code=400,
    )
    with raises(ClientError) as err:
        meetings.update_room(room_id='b3142c46-d1c1-4405-baa6-85683827ed69', params={})
    assert (
        str(err.value)
        == 'Status Code 400: BadRequestError: The room with id: b3142c46-d1c1-4405-baa6-85683827ed69 could not be updated because of its type: temporary'
    )


@responses.activate
def test_update_room_error_no_params_specified(meetings):
    stub(
        responses.PATCH,
        'https://api-eu.vonage.com/beta/meetings/rooms/33791484-231c-421b-8349-96e1a44e27d2',
        fixture_path='meetings/update_room_type_error.json',
        status_code=400,
    )
    with raises(TypeError) as err:
        meetings.update_room(room_id='33791484-231c-421b-8349-96e1a44e27d2')
    assert "update_room() missing 1 required positional argument: 'params'" in str(err.value)


@responses.activate
def test_get_recording(meetings):
    stub(
        responses.GET,
        'https://api-eu.vonage.com/beta/meetings/recordings/e5b73c98-c087-4ee5-b61b-0ea08204fc65',
        fixture_path='meetings/get_recording.json',
    )

    recording = meetings.get_recording(recording_id='e5b73c98-c087-4ee5-b61b-0ea08204fc65')
    assert (
        recording['session_id']
        == '1_MX40NjMzOTg5Mn5-MTY3NDYxNDI4NjY5M35WM0xaVXBSc1lpT3hKWE1XQ2diM1B3cXB-fn4'
    )
    assert recording['started_at'] == '2023-01-25T02:38:31.000Z'
    assert recording['status'] == 'uploaded'


@responses.activate
def test_get_recording_not_found(meetings):
    stub(
        responses.GET,
        'https://api-eu.vonage.com/beta/meetings/recordings/not-a-real-recording-id',
        fixture_path='meetings/get_recording_not_found.json',
        status_code=404,
    )

    with raises(ClientError) as err:
        meetings.get_recording(recording_id='not-a-real-recording-id')
    assert (
        str(err.value)
        == 'Status Code 404: NotFoundError: Recording not-a-real-recording-id was not found'
    )


@responses.activate
def test_delete_recording(meetings):
    stub(
        responses.DELETE,
        'https://api-eu.vonage.com/beta/meetings/recordings/e5b73c98-c087-4ee5-b61b-0ea08204fc65',
        fixture_path='no_content.json',
    )

    assert meetings.delete_recording(recording_id='e5b73c98-c087-4ee5-b61b-0ea08204fc65') == None


@responses.activate
def test_delete_recording_not_uploaded(meetings, client):
    stub(
        responses.DELETE,
        'https://api-eu.vonage.com/beta/meetings/recordings/881f0dbe-3d91-4fd6-aeea-0eca4209b512',
        fixture_path='meetings/delete_recording_not_found.json',
        status_code=404,
    )

    with raises(ClientError) as err:
        meetings.delete_recording(recording_id='881f0dbe-3d91-4fd6-aeea-0eca4209b512')
    assert str(err.value) == 'Status Code 404: NotFoundError: Could not find recording'


@responses.activate
def test_get_session_recordings(meetings):
    stub(
        responses.GET,
        'https://api-eu.vonage.com/beta/meetings/sessions/1_MX40NjMzOTg5Mn5-MTY3NDYxNDI4NjY5M35WM0xaVXBSc1lpT3hKWE1XQ2diM1B3cXB-fn4/recordings',
        fixture_path='meetings/get_session_recordings.json',
    )

    session = meetings.get_session_recordings(
        session_id='1_MX40NjMzOTg5Mn5-MTY3NDYxNDI4NjY5M35WM0xaVXBSc1lpT3hKWE1XQ2diM1B3cXB-fn4'
    )
    assert session['_embedded']['recordings'][0]['id'] == 'e5b73c98-c087-4ee5-b61b-0ea08204fc65'
    assert session['_embedded']['recordings'][0]['started_at'] == '2023-01-25T02:38:31.000Z'
    assert session['_embedded']['recordings'][0]['status'] == 'uploaded'


@responses.activate
def test_get_session_recordings_not_found(meetings):
    stub(
        responses.GET,
        'https://api-eu.vonage.com/beta/meetings/sessions/not-a-real-session-id/recordings',
        fixture_path='meetings/get_session_recordings_not_found.json',
        status_code=404,
    )

    with raises(ClientError) as err:
        meetings.get_session_recordings(session_id='not-a-real-session-id')
    assert (
        str(err.value)
        == 'Status Code 404: NotFoundError: Failed to find session recordings by id: not-a-real-session-id'
    )


@responses.activate
def test_list_dial_in_numbers(meetings):
    stub(
        responses.GET,
        'https://api-eu.vonage.com/beta/meetings/dial-in-numbers',
        fixture_path='meetings/list_dial_in_numbers.json',
    )

    numbers = meetings.list_dial_in_numbers()
    assert numbers[0]['number'] == '541139862166'
    assert numbers[0]['display_name'] == 'Argentina'
    assert numbers[1]['number'] == '442381924626'
    assert numbers[1]['locale'] == 'en-GB'


@responses.activate
def test_list_themes(meetings):
    stub(
        responses.GET,
        'https://api-eu.vonage.com/beta/meetings/themes',
        fixture_path='meetings/list_themes.json',
    )

    themes = meetings.list_themes()
    assert themes[0]['theme_id'] == '1fc39568-bc50-464f-82dc-01e13bed0908'
    assert themes[0]['main_color'] == '#FF0000'
    assert themes[0]['brand_text'] == 'My Other Company'
    assert themes[1]['theme_id'] == '90a21428-b74a-4221-adc3-783935d654db'
    assert themes[1]['main_color'] == '#12f64e'
    assert themes[1]['brand_text'] == 'My Company'


@responses.activate
def test_list_themes_no_themes(meetings):
    stub(
        responses.GET,
        'https://api-eu.vonage.com/beta/meetings/themes',
        fixture_path='meetings/empty_themes.json',
    )

    assert meetings.list_themes() == {}


@responses.activate
def test_create_theme(meetings):
    stub(
        responses.POST,
        "https://api-eu.vonage.com/beta/meetings/themes",
        fixture_path='meetings/theme.json',
    )

    params = {
        'theme_name': 'my_theme',
        'main_color': '#12f64e',
        'brand_text': 'My Company',
        'short_company_url': 'my-company',
    }

    theme = meetings.create_theme(params)
    assert theme['theme_id'] == '90a21428-b74a-4221-adc3-783935d654db'
    assert theme['main_color'] == '#12f64e'
    assert theme['brand_text'] == 'My Company'
    assert theme['domain'] == 'VCP'


def test_create_theme_missing_required_params(meetings):
    with raises(MeetingsError) as err:
        meetings.create_theme({})
    assert str(err.value) == 'Values for "main_color" and "brand_text" must be specified'


@responses.activate
def test_create_theme_name_already_in_use(meetings):
    stub(
        responses.POST,
        "https://api-eu.vonage.com/beta/meetings/themes",
        fixture_path='meetings/theme_name_in_use.json',
        status_code=409,
    )

    params = {
        'theme_name': 'my_theme',
        'main_color': '#12f64e',
        'brand_text': 'My Company',
    }

    with raises(ClientError) as err:
        meetings.create_theme(params)
    assert (
        str(err.value) == 'Status Code 409: ConflictError: theme_name already exists in application'
    )


@responses.activate
def test_get_theme(meetings):
    stub(
        responses.GET,
        "https://api-eu.vonage.com/beta/meetings/themes/90a21428-b74a-4221-adc3-783935d654db",
        fixture_path='meetings/theme.json',
    )

    theme = meetings.get_theme('90a21428-b74a-4221-adc3-783935d654db')
    assert theme['main_color'] == '#12f64e'
    assert theme['brand_text'] == 'My Company'


@responses.activate
def test_get_theme_not_found(meetings):
    stub(
        responses.GET,
        "https://api-eu.vonage.com/beta/meetings/themes/90a21428-b74a-4221-adc3-783935d654dc",
        fixture_path='meetings/theme_not_found.json',
        status_code=404,
    )

    with raises(ClientError) as err:
        meetings.get_theme('90a21428-b74a-4221-adc3-783935d654dc')
    assert (
        str(err.value)
        == 'Status Code 404: NotFoundError: could not find theme 90a21428-b74a-4221-adc3-783935d654dc'
    )


@responses.activate
def test_delete_theme(meetings):
    stub(
        responses.DELETE,
        "https://api-eu.vonage.com/beta/meetings/themes/90a21428-b74a-4221-adc3-783935d654db",
        fixture_path='no_content.json',
    )

    theme = meetings.delete_theme('90a21428-b74a-4221-adc3-783935d654db')
    assert theme == None


@responses.activate
def test_delete_theme_not_found(meetings):
    stub(
        responses.DELETE,
        "https://api-eu.vonage.com/beta/meetings/themes/90a21428-b74a-4221-adc3-783935d654dc",
        fixture_path='meetings/theme_not_found.json',
        status_code=404,
    )

    with raises(ClientError) as err:
        meetings.delete_theme('90a21428-b74a-4221-adc3-783935d654dc')
    assert (
        str(err.value)
        == 'Status Code 404: NotFoundError: could not find theme 90a21428-b74a-4221-adc3-783935d654dc'
    )


@responses.activate
def test_delete_theme_in_use(meetings):
    stub(
        responses.DELETE,
        "https://api-eu.vonage.com/beta/meetings/themes/90a21428-b74a-4221-adc3-783935d654db",
        fixture_path='meetings/delete_theme_in_use.json',
        status_code=400,
    )

    with raises(ClientError) as err:
        meetings.delete_theme('90a21428-b74a-4221-adc3-783935d654db')
    assert (
        str(err.value)
        == 'Status Code 400: BadRequestError: could not delete theme\nError: Theme 90a21428-b74a-4221-adc3-783935d654db is used by 1 room'
    )


@responses.activate
def test_update_theme(meetings):
    stub(
        responses.PATCH,
        "https://api-eu.vonage.com/beta/meetings/themes/90a21428-b74a-4221-adc3-783935d654db",
        fixture_path='meetings/updated_theme.json',
    )

    params = {
        'update_details': {
            'theme_name': 'updated_theme',
            'main_color': '#FF0000',
            'brand_text': 'My Updated Company Name',
            'short_company_url': 'updated_company_url',
        }
    }

    theme = meetings.update_theme('90a21428-b74a-4221-adc3-783935d654db', params)
    assert theme['theme_id'] == '90a21428-b74a-4221-adc3-783935d654db'
    assert theme['main_color'] == '#FF0000'
    assert theme['brand_text'] == 'My Updated Company Name'
    assert theme['short_company_url'] == 'updated_company_url'


@responses.activate
def test_update_theme_no_keys(meetings):
    stub(
        responses.PATCH,
        "https://api-eu.vonage.com/beta/meetings/themes/90a21428-b74a-4221-adc3-783935d654db",
        fixture_path='meetings/update_no_keys.json',
        status_code=400,
    )

    with raises(ClientError) as err:
        meetings.update_theme('90a21428-b74a-4221-adc3-783935d654db', {'update_details': {}})
    assert (
        str(err.value)
        == 'Status Code 400: InputValidationError: "update_details" must have at least 1 key'
    )


@responses.activate
def test_update_theme_not_found(meetings):
    stub(
        responses.PATCH,
        "https://api-eu.vonage.com/beta/meetings/themes/90a21428-b74a-4221-adc3-783935d654dc",
        fixture_path='meetings/theme_not_found.json',
        status_code=404,
    )

    with raises(ClientError) as err:
        meetings.update_theme(
            '90a21428-b74a-4221-adc3-783935d654dc',
            {'update_details': {'theme_name': 'my_new_name'}},
        )
    assert (
        str(err.value)
        == 'Status Code 404: NotFoundError: could not find theme 90a21428-b74a-4221-adc3-783935d654dc'
    )


@responses.activate
def test_update_theme_name_already_exists(meetings):
    stub(
        responses.PATCH,
        'https://api-eu.vonage.com/beta/meetings/themes/90a21428-b74a-4221-adc3-783935d654db',
        fixture_path='meetings/update_theme_already_exists.json',
        status_code=409,
    )

    with raises(ClientError) as err:
        meetings.update_theme(
            '90a21428-b74a-4221-adc3-783935d654db',
            {'update_details': {'theme_name': 'my_other_theme'}},
        )
    assert (
        str(err.value) == 'Status Code 409: ConflictError: theme_name already exists in application'
    )


@responses.activate
def test_list_rooms_with_options(meetings):
    stub(
        responses.GET,
        'https://api-eu.vonage.com/beta/meetings/themes/90a21428-b74a-4221-adc3-783935d654db/rooms',
        fixture_path='meetings/list_rooms_with_theme_id.json',
    )

    rooms = meetings.list_rooms_with_theme_id(
        '90a21428-b74a-4221-adc3-783935d654db',
        page_size=5,
        start_id=0,
        end_id=99999999,
    )
    assert rooms['_embedded'][0]['id'] == '33791484-231c-421b-8349-96e1a44e27d2'
    assert rooms['_embedded'][0]['display_name'] == 'test_long_term_room'
    assert rooms['_embedded'][0]['theme_id'] == '90a21428-b74a-4221-adc3-783935d654db'
    assert rooms['page_size'] == 5
    assert (
        rooms['_links']['self']['href']
        == 'api-eu.vonage.com/meetings/rooms?page_size=20&start_id=2009870'
    )


@responses.activate
def test_list_rooms_with_theme_id_not_found(meetings):
    stub(
        responses.GET,
        'https://api-eu.vonage.com/beta/meetings/themes/90a21428-b74a-4221-adc3-783935d654dc/rooms',
        fixture_path='meetings/list_rooms_theme_id_not_found.json',
        status_code=404,
    )

    with raises(ClientError) as err:
        meetings.list_rooms_with_theme_id(
            '90a21428-b74a-4221-adc3-783935d654dc', start_id=0, end_id=99999999
        )
    assert (
        str(err.value)
        == 'Status Code 404: NotFoundError: Failed to get rooms because theme id 90a21428-b74a-4221-adc3-783935d654dc not found'
    )


@responses.activate
def test_update_application_theme(meetings):
    stub(
        responses.PATCH,
        'https://api-eu.vonage.com/beta/meetings/applications',
        fixture_path='meetings/update_application_theme.json',
    )

    response = meetings.update_application_theme(theme_id='90a21428-b74a-4221-adc3-783935d654db')
    assert response['application_id'] == 'my-application-id'
    assert response['account_id'] == 'my-account-id'
    assert response['default_theme_id'] == '90a21428-b74a-4221-adc3-783935d654db'


@responses.activate
def test_update_application_theme_bad_request(meetings):
    stub(
        responses.PATCH,
        'https://api-eu.vonage.com/beta/meetings/applications',
        fixture_path='meetings/update_application_theme_id_not_found.json',
        status_code=400,
    )

    with raises(ClientError) as err:
        meetings.update_application_theme(theme_id='not-a-real-theme-id')
    assert (
        str(err.value)
        == 'Status Code 400: BadRequestError: Failed to update application because theme id not-a-real-theme-id not found'
    )


@responses.activate
def test_upload_logo_to_theme(meetings):
    stub(
        responses.GET,
        'https://api-eu.vonage.com/beta/meetings/themes/logos-upload-urls',
        fixture_path='meetings/list_logo_upload_urls.json',
    )
    stub(
        responses.POST,
        'https://s3.amazonaws.com/roomservice-whitelabel-logos-prod',
        fixture_path='no_content.json',
        status_code=204,
    )
    stub_bytes(
        responses.PUT,
        'https://api-eu.vonage.com/beta/meetings/themes/90a21428-b74a-4221-adc3-783935d654db/finalizeLogos',
        body=b'OK',
    )

    response = meetings.upload_logo_to_theme(
        theme_id='90a21428-b74a-4221-adc3-783935d654db',
        path_to_image='tests/data/meetings/transparent_logo.png',
        logo_type='white',
    )
    assert response == 'Logo upload to theme: 90a21428-b74a-4221-adc3-783935d654db was successful.'


@responses.activate
def test_get_logo_upload_url(meetings):
    stub(
        responses.GET,
        'https://api-eu.vonage.com/beta/meetings/themes/logos-upload-urls',
        fixture_path='meetings/list_logo_upload_urls.json',
    )

    url_w = meetings._get_logo_upload_url('white')
    assert url_w['url'] == 'https://s3.amazonaws.com/roomservice-whitelabel-logos-prod'
    assert url_w['fields']['X-Amz-Credential'] == 'some-credential'
    assert (
        url_w['fields']['key']
        == 'auto-expiring-temp/logos/white/d92b31ae-fbf1-4709-a729-c0fa75368c25'
    )
    assert url_w['fields']['logoType'] == 'white'
    url_c = meetings._get_logo_upload_url('colored')
    assert (
        url_c['fields']['key']
        == 'auto-expiring-temp/logos/colored/c4e00bac-781b-4bf0-bd5f-b9ff2cbc1b6c'
    )
    assert url_c['fields']['logoType'] == 'colored'
    url_f = meetings._get_logo_upload_url('favicon')
    assert (
        url_f['fields']['key']
        == 'auto-expiring-temp/logos/favicon/d7a81477-38f7-460c-b51f-1462b8426df5'
    )
    assert url_f['fields']['logoType'] == 'favicon'

    with raises(MeetingsError) as err:
        meetings._get_logo_upload_url('not-a-valid-option')
    assert str(err.value) == 'Cannot find the upload URL for the specified logo type.'


@responses.activate
def test_upload_to_aws(meetings):
    stub(
        responses.POST,
        'https://s3.amazonaws.com/roomservice-whitelabel-logos-prod',
        fixture_path='no_content.json',
        status_code=204,
    )

    with open('tests/data/meetings/list_logo_upload_urls.json') as file:
        urls = json.load(file)
    params = urls[0]
    meetings._upload_to_aws(params, 'tests/data/meetings/transparent_logo.png')


@responses.activate
def test_upload_to_aws_error(meetings):
    stub(
        responses.POST,
        'https://s3.amazonaws.com/not-a-valid-url',
        status_code=403,
        fixture_path='meetings/upload_to_aws_error.xml',
    )

    with open('tests/data/meetings/list_logo_upload_urls.json') as file:
        urls = json.load(file)

    params = urls[0]
    params['url'] = 'https://s3.amazonaws.com/not-a-valid-url'
    with raises(MeetingsError) as err:
        meetings._upload_to_aws(params, 'tests/data/meetings/transparent_logo.png')
    assert (
        str(err.value)
        == 'Logo upload process failed. b\'<?xml version="1.0" encoding="UTF-8"?>\\\\n<Error><Code>SignatureDoesNotMatch</Code><Message>The request signature we calculated does not match the signature you provided. Check your key and signing method.</Message><AWSAccessKeyId>ASIA5NAYMMB6M7A2QEAR</AWSAccessKeyId><StringToSign></StringToSign><SignatureProvided>b2f311449e26692a174ab2c7ca2afab24bd19c509cc611a4cef7cb2c5bb2ea9a</SignatureProvided><StringToSignBytes></StringToSignBytes><RequestId>5ZS7MSFN46X89NXA</RequestId><HostId>f+HV7uSpeawLv5lFvN+QiYP6swbiTMd/XaJeVGC+/pqKHlwlgKZ6vg+qBjV/ufb1e5WS/bxBM/Y=</HostId></Error>\''
    )


@responses.activate
def test_add_logo_to_theme(meetings):
    stub_bytes(
        responses.PUT,
        'https://api-eu.vonage.com/beta/meetings/themes/90a21428-b74a-4221-adc3-783935d654db/finalizeLogos',
        body=b'OK',
    )

    response = meetings._add_logo_to_theme(
        theme_id='90a21428-b74a-4221-adc3-783935d654db',
        key='auto-expiring-temp/logos/white/d92b31ae-fbf1-4709-a729-c0fa75368c25',
    )
    assert response == b'OK'


@responses.activate
def test_add_logo_to_theme_key_error(meetings):
    stub(
        responses.PUT,
        'https://api-eu.vonage.com/beta/meetings/themes/90a21428-b74a-4221-adc3-783935d654dc/finalizeLogos',
        fixture_path='meetings/logo_key_error.json',
        status_code=400,
    )

    with raises(ClientError) as err:
        meetings._add_logo_to_theme(
            theme_id='90a21428-b74a-4221-adc3-783935d654dc',
            key='an-invalid-key',
        )
    assert (
        str(err.value)
        == "Status Code 400: BadRequestError: could not finalize logos\nError: {'logoKey': 'not-a-key', 'code': 'key_not_found'}"
    )


@responses.activate
def test_add_logo_to_theme_not_found_error(meetings):
    stub(
        responses.PUT,
        'https://api-eu.vonage.com/beta/meetings/themes/90a21428-b74a-4221-adc3-783935d654dc/finalizeLogos',
        fixture_path='meetings/theme_not_found.json',
        status_code=404,
    )

    with raises(ClientError) as err:
        meetings._add_logo_to_theme(
            theme_id='90a21428-b74a-4221-adc3-783935d654dc',
            key='auto-expiring-temp/logos/white/d92b31ae-fbf1-4709-a729-c0fa75368c25',
        )
    assert (
        str(err.value)
        == 'Status Code 404: NotFoundError: could not find theme 90a21428-b74a-4221-adc3-783935d654dc'
    )



from util import *


@responses.activate
def test_send_2fa_message(short_codes, dummy_data):
    stub(responses.POST, "https://rest.nexmo.com/sc/us/2fa/json")

    params = {"to": "16365553226", "pin": "1234"}

    assert isinstance(short_codes.send_2fa_message(params), dict)
    assert request_user_agent() == dummy_data.user_agent
    assert "to=16365553226" in request_body()
    assert "pin=1234" in request_body()


@responses.activate
def test_send_event_alert_message(short_codes, dummy_data):
    stub(responses.POST, "https://rest.nexmo.com/sc/us/alert/json")

    params = {"to": "16365553226", "server": "host", "link": "http://example.com/"}

    assert isinstance(short_codes.send_event_alert_message(params), dict)
    assert request_user_agent() == dummy_data.user_agent
    assert "to=16365553226" in request_body()
    assert "server=host" in request_body()
    assert "link=http%3A%2F%2Fexample.com%2F" in request_body()


@responses.activate
def test_send_marketing_message(short_codes, dummy_data):
    stub(responses.POST, "https://rest.nexmo.com/sc/us/marketing/json")

    params = {
        "from": "short-code",
        "to": "16365553226",
        "keyword": "NEXMO",
        "text": "Hello",
    }

    assert isinstance(short_codes.send_marketing_message(params), dict)
    assert request_user_agent() == dummy_data.user_agent
    assert "from=short-code" in request_body()
    assert "to=16365553226" in request_body()
    assert "keyword=NEXMO" in request_body()
    assert "text=Hello" in request_body()


@responses.activate
def test_get_event_alert_numbers(short_codes, dummy_data):
    stub(responses.GET, "https://rest.nexmo.com/sc/us/alert/opt-in/query/json")

    assert isinstance(short_codes.get_event_alert_numbers(), dict)
    assert request_user_agent() == dummy_data.user_agent


@responses.activate
def test_resubscribe_event_alert_number(short_codes, dummy_data):
    stub(responses.POST, "https://rest.nexmo.com/sc/us/alert/opt-in/manage/json")

    params = {"msisdn": "441632960960"}

    assert isinstance(short_codes.resubscribe_event_alert_number(params), dict)
    assert request_user_agent() == dummy_data.user_agent
    assert "msisdn=441632960960" in request_body()


import vonage
from vonage_jwt.jwt import JwtClient

from .account import Account
from .application import ApplicationV2, Application
from .errors import *
from .meetings import Meetings
from .messages import Messages
from .number_insight import NumberInsight
from .number_management import Numbers
from .proactive_connect import ProactiveConnect
from .redact import Redact
from .short_codes import ShortCodes
from .sms import Sms
from .subaccounts import Subaccounts
from .users import Users
from .ussd import Ussd
from .voice import Voice
from .verify import Verify
from .verify2 import Verify2

import logging
from platform import python_version

import base64
import hashlib
import hmac
import os
import time

from requests import Response
from requests.adapters import HTTPAdapter
from requests.sessions import Session

string_types = (str, bytes)

try:
    from json import JSONDecodeError
except ImportError:
    JSONDecodeError = ValueError

logger = logging.getLogger("vonage")


class Client:
    """
    Create a Client object to start making calls to Vonage/Nexmo APIs.

    The credentials you provide when instantiating a Client determine which
    methods can be called. Consult the `Vonage API docs <https://developer.vonage.com/concepts/guides/authentication/>`
    for details of the authentication used by the APIs you wish to use, and instantiate your
    client with the appropriate credentials.

    :param str key: Your Vonage API key
    :param str secret: Your Vonage API secret.
    :param str signature_secret: Your Vonage API signature secret.
        You may need to have this enabled by Vonage support. It is only used for SMS authentication.
    :param str signature_method:
        The encryption method used for signature encryption. This must match the method
        configured in the Vonage Dashboard. We recommend `sha256` or `sha512`.
        This should be one of `md5`, `sha1`, `sha256`, or `sha512` if using HMAC digests.
        If you want to use a simple MD5 hash, leave this as `None`.
    :param str application_id: Your application ID if calling methods which use JWT authentication.
    :param str private_key: Your private key, for calling methods which use JWT authentication.
        This should either be a str containing the key in its PEM form, or a path to a private key file.
    :param str app_name: This optional value is added to the user-agent header
        provided by this library and can be used to track your app statistics.
    :param str app_version: This optional value is added to the user-agent header
        provided by this library and can be used to track your app statistics.
    :param timeout: (optional) How many seconds to wait for the server to send data
        before giving up, as a float, or a (connect timeout, read
        timeout) tuple. If set this timeout is used for every call to the Vonage enpoints
    :type timeout: float or tuple
    """

    def __init__(
        self,
        key=None,
        secret=None,
        signature_secret=None,
        signature_method=None,
        application_id=None,
        private_key=None,
        app_name=None,
        app_version=None,
        timeout=None,
        pool_connections=10,
        pool_maxsize=10,
        max_retries=3,
    ):
        self.api_key = key or os.environ.get("VONAGE_API_KEY", None)
        self.api_secret = secret or os.environ.get("VONAGE_API_SECRET", None)

        self.signature_secret = signature_secret or os.environ.get("VONAGE_SIGNATURE_SECRET", None)
        self.signature_method = signature_method or os.environ.get("VONAGE_SIGNATURE_METHOD", None)

        if self.signature_method in {"md5", "sha1", "sha256", "sha512"}:
            self.signature_method = getattr(hashlib, signature_method)

        if private_key is not None and application_id is not None:
            self._jwt_client = JwtClient(application_id, private_key)

        self._jwt_claims = {}
        self._host = "rest.nexmo.com"
        self._api_host = "api.nexmo.com"
        self._meetings_api_host = "api-eu.vonage.com/beta/meetings"
        self._proactive_connect_host = "api-eu.vonage.com"

        user_agent = f"vonage-python/{vonage.__version__} python/{python_version()}"

        if app_name and app_version:
            user_agent += f" {app_name}/{app_version}"

        self.headers = {"User-Agent": user_agent, "Accept": "application/json"}

        self.account = Account(self)
        self.application = Application(self)
        self.meetings = Meetings(self)
        self.messages = Messages(self)
        self.number_insight = NumberInsight(self)
        self.numbers = Numbers(self)
        self.proactive_connect = ProactiveConnect(self)
        self.short_codes = ShortCodes(self)
        self.sms = Sms(self)
        self.subaccounts = Subaccounts(self)
        self.users = Users(self)
        self.ussd = Ussd(self)
        self.verify = Verify(self)
        self.verify2 = Verify2(self)
        self.voice = Voice(self)

        self.timeout = timeout
        self.session = Session()
        self.adapter = HTTPAdapter(
            pool_connections=pool_connections, pool_maxsize=pool_maxsize, max_retries=max_retries
        )
        self.session.mount("https://", self.adapter)

    # Gets and sets _host attribute
    def host(self, value=None):
        if value is None:
            return self._host
        else:
            self._host = value

    # Gets and sets _api_host attribute
    def api_host(self, value=None):
        if value is None:
            return self._api_host
        else:
            self._api_host = value

    # Gets and sets _meetings_api_host attribute
    def meetings_api_host(self, value=None):
        if value is None:
            return self._meetings_api_host
        else:
            self._meetings_api_host = value

    def proactive_connect_host(self, value=None):
        if value is None:
            return self._proactive_connect_host
        else:
            self._proactive_connect_host = value

    def auth(self, params=None, **kwargs):
        self._jwt_claims = params or kwargs

    def check_signature(self, params):
        params = dict(params)
        signature = params.pop("sig", "").lower()
        return hmac.compare_digest(signature, self.signature(params))

    def signature(self, params):
        if self.signature_method:
            hasher = hmac.new(self.signature_secret.encode(), digestmod=self.signature_method)
        else:
            hasher = hashlib.md5()

        # Add timestamp if not already present
        if not params.get("timestamp"):
            params["timestamp"] = int(time.time())

        for key in sorted(params):
            value = params[key]

            if isinstance(value, str):
                value = value.replace("&", "_").replace("=", "_")

            hasher.update(f"&{key}={value}".encode("utf-8"))

        if self.signature_method is None:
            hasher.update(self.signature_secret.encode())

        return hasher.hexdigest()

    def get(self, host, request_uri, params=None, auth_type=None):
        uri = f"https://{host}{request_uri}"
        self._request_headers = self.headers

        if auth_type == 'jwt':
            self._request_headers['Authorization'] = self._create_jwt_auth_string()
        elif auth_type == 'params':
            params = dict(params or {}, api_key=self.api_key, api_secret=self.api_secret)
        elif auth_type == 'header':
            self._request_headers['Authorization'] = self._create_header_auth_string()
        else:
            raise InvalidAuthenticationTypeError(
                f'Invalid authentication type. Must be one of "jwt", "header" or "params".'
            )

        logger.debug(
            f"GET to {repr(uri)} with params {repr(params)}, headers {repr(self._request_headers)}"
        )
        return self.parse(
            host,
            self.session.get(
                uri, params=params, headers=self._request_headers, timeout=self.timeout
            ),
        )

    def post(
        self,
        host,
        request_uri,
        params,
        auth_type=None,
        body_is_json=True,
        supports_signature_auth=False,
    ):
        """
        Low-level method to make a post request to an API server.
        This method automatically adds authentication, picking the first applicable authentication method from the following:
        - If the supports_signature_auth param is True, and the client was instantiated with a signature_secret,
            then signature authentication will be used.
        :param bool supports_signature_auth: Preferentially use signature authentication if a signature_secret was provided
            when initializing this client.
        """
        uri = f"https://{host}{request_uri}"
        self._request_headers = self.headers

        if supports_signature_auth and self.signature_secret:
            params["api_key"] = self.api_key
            params["sig"] = self.signature(params)
        elif auth_type == 'jwt':
            self._request_headers['Authorization'] = self._create_jwt_auth_string()
        elif auth_type == 'params':
            params = dict(params, api_key=self.api_key, api_secret=self.api_secret)
        elif auth_type == 'header':
            self._request_headers['Authorization'] = self._create_header_auth_string()
        else:
            raise InvalidAuthenticationTypeError(
                f'Invalid authentication type. Must be one of "jwt", "header" or "params".'
            )

        logger.debug(
            f"POST to {repr(uri)} with params {repr(params)}, headers {repr(self._request_headers)}"
        )
        if body_is_json:
            return self.parse(
                host,
                self.session.post(
                    uri, json=params, headers=self._request_headers, timeout=self.timeout
                ),
            )
        else:
            return self.parse(
                host,
                self.session.post(
                    uri, data=params, headers=self._request_headers, timeout=self.timeout
                ),
            )

    def put(self, host, request_uri, params, auth_type=None):
        uri = f"https://{host}{request_uri}"
        self._request_headers = self.headers

        if auth_type == 'jwt':
            self._request_headers['Authorization'] = self._create_jwt_auth_string()
        elif auth_type == 'header':
            self._request_headers['Authorization'] = self._create_header_auth_string()
        else:
            raise InvalidAuthenticationTypeError(
                f'Invalid authentication type. Must be one of "jwt", "header" or "params".'
            )

        logger.debug(
            f"PUT to {repr(uri)} with params {repr(params)}, headers {repr(self._request_headers)}"
        )
        # All APIs that currently use put methods require a json-formatted body so don't need to check this
        return self.parse(
            host,
            self.session.put(uri, json=params, headers=self._request_headers, timeout=self.timeout),
        )

    def patch(self, host, request_uri, params, auth_type=None):
        uri = f"https://{host}{request_uri}"
        self._request_headers = self.headers

        if auth_type == 'jwt':
            self._request_headers['Authorization'] = self._create_jwt_auth_string()
        elif auth_type == 'header':
            self._request_headers['Authorization'] = self._create_header_auth_string()
        else:
            raise InvalidAuthenticationTypeError(f"""Invalid authentication type.""")

        logger.debug(
            f"PATCH to {repr(uri)} with params {repr(params)}, headers {repr(self._request_headers)}"
        )
        # Only newer APIs (that expect json-bodies) currently use this method, so we will always send a json-formatted body
        return self.parse(host, self.session.patch(uri, json=params, headers=self._request_headers))

    def delete(self, host, request_uri, params=None, auth_type=None):
        uri = f"https://{host}{request_uri}"
        self._request_headers = self.headers

        if auth_type == 'jwt':
            self._request_headers['Authorization'] = self._create_jwt_auth_string()
        elif auth_type == 'header':
            self._request_headers['Authorization'] = self._create_header_auth_string()
        else:
            raise InvalidAuthenticationTypeError(
                f'Invalid authentication type. Must be one of "jwt", "header" or "params".'
            )

        logger.debug(f"DELETE to {repr(uri)} with headers {repr(self._request_headers)}")
        if params is not None:
            logger.debug(f"DELETE call has params {repr(params)}")
        return self.parse(
            host,
            self.session.delete(
                uri, headers=self._request_headers, timeout=self.timeout, params=params
            ),
        )

    def parse(self, host, response: Response):
        logger.debug(f"Response headers {repr(response.headers)}")
        if response.status_code == 401:
            raise AuthenticationError("Authentication failed.")
        elif response.status_code == 204:
            return None
        elif 200 <= response.status_code < 300:
            # Strip off any encoding from the content-type header:
            try:
                content_mime = response.headers.get("content-type").split(";", 1)[0]
            except AttributeError:
                if response.json() is None:
                    return None
            if content_mime == "application/json":
                try:
                    return response.json()
                except JSONDecodeError:
                    pass
            else:
                return response.content
        elif 400 <= response.status_code < 500:
            logger.warning(f"Client error: {response.status_code} {repr(response.content)}")
            message = f"{response.status_code} response from {host}"

            # Test for standard error format:
            try:
                error_data = response.json()
                if "type" in error_data and "title" in error_data and "detail" in error_data:
                    title = error_data["title"]
                    detail = error_data["detail"]
                    type = error_data["type"]
                    message = f"{title}: {detail} ({type}){self._add_individual_errors(error_data)}"
                elif 'status' in error_data and 'message' in error_data and 'name' in error_data:
                    message = (
                        f'Status Code {error_data["status"]}: {error_data["name"]}: {error_data["message"]}'
                        f'{self._add_individual_errors(error_data)}'
                    )
                else:
                    message = error_data
            except JSONDecodeError:
                pass
            raise ClientError(message)

        elif 500 <= response.status_code < 600:
            logger.warning(f"Server error: {response.status_code} {repr(response.content)}")
            message = f"{response.status_code} response from {host}"
            raise ServerError(message)

    def _add_individual_errors(self, error_data):
        message = ''
        if 'errors' in error_data:
            for error in error_data["errors"]:
                message += f"\nError: {error}"
        return message

    def _create_jwt_auth_string(self):
        return b"Bearer " + self._generate_application_jwt()

    def _generate_application_jwt(self):
        try:
            return self._jwt_client.generate_application_jwt(self._jwt_claims)
        except AttributeError as err:
            if '_jwt_client' in str(err):
                raise ClientError(
                    'JWT generation failed. Check that you passed in valid values for "application_id" and "private_key".'
                )
            else:
                raise err

    def _create_header_auth_string(self):
        hash = base64.b64encode(f"{self.api_key}:{self.api_secret}".encode("utf-8")).decode("ascii")
        return f"Basic {hash}"



from pydantic import BaseModel, Field, validator, constr, confloat, conint
from typing import Optional, Union, List
from typing_extensions import Literal

from .connect_endpoints import ConnectEndpoints
from .input_types import InputTypes
from .pay_prompts import PayPrompts

from deprecated import deprecated


class Ncco:
    class Action(BaseModel):
        action: str = None

    class Record(Action):
        """Use the record action to record a call or part of a call."""

        action = Field('record', const=True)
        format: Optional[Literal['mp3', 'wav', 'ogg']]
        split: Optional[Literal['conversation']]
        channels: Optional[conint(ge=1, le=32)]
        endOnSilence: Optional[conint(ge=3, le=10)]
        endOnKey: Optional[constr(regex='^[0-9*#]$')]
        timeOut: Optional[conint(ge=3, le=7200)]
        beepStart: Optional[bool]
        eventUrl: Optional[Union[List[str], str]]
        eventMethod: Optional[constr(to_upper=True)]

        @validator('channels')
        def enable_split(cls, v, values):
            if values['split'] is None:
                values['split'] = 'conversation'
            return v

        @validator('eventUrl')
        def ensure_url_in_list(cls, v):
            return Ncco._ensure_object_in_list(v)

    class Conversation(Action):
        """You can use the conversation action to create standard or moderated conferences,
        while preserving the communication context.
        Using conversation with the same name reuses the same persisted conversation."""

        action = Field('conversation', const=True)
        name: str
        musicOnHoldUrl: Optional[Union[List[str], str]]
        startOnEnter: Optional[bool]
        endOnExit: Optional[bool]
        record: Optional[bool]
        canSpeak: Optional[List[str]]
        canHear: Optional[List[str]]
        mute: Optional[bool]

        @validator('musicOnHoldUrl')
        def ensure_url_in_list(cls, v):
            return Ncco._ensure_object_in_list(v)

        @validator('mute')
        def can_mute(cls, v, values):
            if 'canSpeak' in values and values['canSpeak'] is not None:
                raise ValueError('Cannot use mute option if canSpeak option is specified.')
            return v

    class Connect(Action):
        """You can use the connect action to connect a call to endpoints such as phone numbers or a VBC extension."""

        action = Field('connect', const=True)
        endpoint: Union[dict, ConnectEndpoints.Endpoint, List[dict]]
        from_: Optional[constr(regex=r'^[1-9]\d{6,14}$')]
        randomFromNumber: Optional[bool]
        eventType: Optional[Literal['synchronous']]
        timeout: Optional[int]
        limit: Optional[conint(le=7200)]
        machineDetection: Optional[Literal['continue', 'hangup']]
        advancedMachineDetection: Optional[dict]
        eventUrl: Optional[Union[List[str], str]]
        eventMethod: Optional[constr(to_upper=True)]
        ringbackTone: Optional[str]

        @validator('endpoint')
        def validate_endpoint(cls, v):
            if type(v) is dict:
                return [ConnectEndpoints.create_endpoint_model_from_dict(v)]
            elif type(v) is list:
                return [ConnectEndpoints.create_endpoint_model_from_dict(v[0])]
            else:
                return [v]

        @validator('from_')
        def set_from_field(cls, v, values):
            values['from'] = v

        @validator('randomFromNumber')
        def check_from_not_set(cls, v, values):
            if v is True and 'from' in values:
                if values['from'] is not None:
                    raise ValueError(
                        'Cannot set a "from" ("from_") field and also the "randomFromNumber" = True option'
                    )
            return v

        @validator('eventUrl')
        def ensure_url_in_list(cls, v):
            return Ncco._ensure_object_in_list(v)

        @validator('advancedMachineDetection')
        def validate_advancedMachineDetection(cls, v):
            if 'behavior' in v and v['behavior'] not in ('continue', 'hangup'):
                raise ValueError('advancedMachineDetection["behavior"] must be one of: "continue", "hangup".')
            if 'mode' in v and v['mode'] not in ('detect, detect_beep'):
                raise ValueError('advancedMachineDetection["mode"] must be one of: "detect", "detect_beep".')
            return v

        class Config:
            smart_union = True

    class Talk(Action):
        """The talk action sends synthesized speech to a Conversation."""

        action = Field('talk', const=True)
        text: constr(max_length=1500)
        bargeIn: Optional[bool]
        loop: Optional[conint(ge=0)]
        level: Optional[confloat(ge=-1, le=1)]
        language: Optional[str]
        style: Optional[int]
        premium: Optional[bool]

    class Stream(Action):
        """The stream action allows you to send an audio stream to a Conversation."""

        action = Field('stream', const=True)
        streamUrl: Union[List[str], str]
        level: Optional[confloat(ge=-1, le=1)]
        bargeIn: Optional[bool]
        loop: Optional[conint(ge=0)]

        @validator('streamUrl')
        def ensure_url_in_list(cls, v):
            return Ncco._ensure_object_in_list(v)

    class Input(Action):
        """Collect digits or speech input by the person you are are calling."""

        action = Field('input', const=True)
        type: Union[
            Literal['dtmf', 'speech'], List[Literal['dtmf']], List[Literal['speech']], List[Literal['dtmf', 'speech']]
        ]
        dtmf: Optional[Union[InputTypes.Dtmf, dict]]
        speech: Optional[Union[InputTypes.Speech, dict]]
        eventUrl: Optional[Union[List[str], str]]
        eventMethod: Optional[constr(to_upper=True)]

        @validator('type', 'eventUrl')
        def ensure_value_in_list(cls, v):
            return Ncco._ensure_object_in_list(v)

        @validator('dtmf')
        def ensure_input_object_is_dtmf_model(cls, v):
            if type(v) is dict:
                return InputTypes.create_dtmf_model(v)
            else:
                return v

        @validator('speech')
        def ensure_input_object_is_speech_model(cls, v):
            if type(v) is dict:
                return InputTypes.create_speech_model(v)
            else:
                return v

    class Notify(Action):
        """Use the notify action to send a custom payload to your event URL."""

        action = Field('notify', const=True)
        payload: dict
        eventUrl: Union[List[str], str]
        eventMethod: Optional[constr(to_upper=True)]

        @validator('eventUrl')
        def ensure_url_in_list(cls, v):
            return Ncco._ensure_object_in_list(v)

    @deprecated(version='3.2.3', reason='The Pay NCCO action has been deprecated.')
    class Pay(Action):
        """The pay action collects credit card information with DTMF input in a secure (PCI-DSS compliant) way."""

        action = Field('pay', const=True)
        amount: confloat(ge=0)
        currency: Optional[constr(to_lower=True)]
        eventUrl: Optional[Union[List[str], str]]
        prompts: Optional[Union[List[PayPrompts.TextPrompt], PayPrompts.TextPrompt, dict]]
        voice: Optional[Union[PayPrompts.VoicePrompt, dict]]

        @validator('amount')
        def round_amount(cls, v):
            return round(v, 2)

        @validator('eventUrl')
        def ensure_url_in_list(cls, v):
            return Ncco._ensure_object_in_list(v)

        @validator('prompts')
        def ensure_text_model(cls, v):
            if type(v) is dict:
                return PayPrompts.create_text_model(v)
            else:
                return v

        @validator('voice')
        def ensure_voice_model(cls, v):
            if type(v) is dict:
                return PayPrompts.create_voice_model(v)
            else:
                return v

    @staticmethod
    def build_ncco(*args: Action, actions: List[Action] = None) -> str:
        ncco = []
        if actions is not None:
            for action in actions:
                ncco.append(action.dict(exclude_none=True))
        for action in args:
            ncco.append(action.dict(exclude_none=True))
        return ncco

    @staticmethod
    def _ensure_object_in_list(obj):
        if type(obj) != list:
            return [obj]
        else:
            return obj



import os
import os.path
import platform

import pytest


# Ensure our client isn't being configured with real values!
os.environ.clear()


def read_file(path):
    with open(os.path.join(os.path.dirname(__file__), path)) as input_file:
        return input_file.read()


class DummyData(object):
    def __init__(self):
        import vonage

        self.api_key = "nexmo-api-key"
        self.api_secret = "nexmo-api-secret"
        self.signature_secret = "secret"
        self.application_id = "nexmo-application-id"
        self.private_key = read_file("data/private_key.txt")
        self.public_key = read_file("data/public_key.txt")
        self.user_agent = f"vonage-python/{vonage.__version__} python/{platform.python_version()}"
        self.host = "rest.nexmo.com"
        self.api_host = "api.nexmo.com"
        self.meetings_api_host = "api-eu.vonage.com/beta/meetings"


@pytest.fixture(scope="session")
def dummy_data():
    return DummyData()


@pytest.fixture
def client(dummy_data):
    import vonage

    return vonage.Client(
        key=dummy_data.api_key,
        secret=dummy_data.api_secret,
        application_id=dummy_data.application_id,
        private_key=dummy_data.private_key,
    )


# Represents an instance of the Voice class for testing
@pytest.fixture
def voice(client):
    import vonage

    return vonage.Voice(client)


# Represents an instance of the Sms class for testing
@pytest.fixture
def sms(client):
    import vonage

    return vonage.Sms(client)


# Represents an instance of the Verify class for testing
@pytest.fixture
def verify(client):
    import vonage

    return vonage.Verify(client)


@pytest.fixture
def number_insight(client):
    import vonage

    return vonage.NumberInsight(client)


@pytest.fixture
def account(client):
    import vonage

    return vonage.Account(client)


@pytest.fixture
def numbers(client):
    import vonage

    return vonage.Numbers(client)


@pytest.fixture
def ussd(client):
    import vonage

    return vonage.Ussd(client)


@pytest.fixture
def short_codes(client):
    import vonage

    return vonage.ShortCodes(client)


@pytest.fixture
def messages(client):
    import vonage

    return vonage.Messages(client)


@pytest.fixture
def redact(client):
    import vonage

    return vonage.Redact(client)


@pytest.fixture
def application_v2(client):
    import vonage

    return vonage.ApplicationV2(client)


@pytest.fixture
def meetings(client):
    import vonage

    return vonage.Meetings(client)


@pytest.fixture
def proc(client):
    import vonage

    return vonage.ProactiveConnect(client)



class Numbers:
    auth_type = 'header'
    defaults = {'auth_type': auth_type, 'body_is_json': False}

    def __init__(self, client):
        self._client = client

    def get_account_numbers(self, params=None, **kwargs):
        return self._client.get(self._client.host(), "/account/numbers", params or kwargs, auth_type=Numbers.auth_type)

    def get_available_numbers(self, country_code, params=None, **kwargs):
        return self._client.get(
            self._client.host(),
            "/number/search",
            dict(params or kwargs, country=country_code),
            auth_type=Numbers.auth_type,
        )

    def buy_number(self, params=None, **kwargs):
        return self._client.post(self._client.host(), "/number/buy", params or kwargs, **Numbers.defaults)

    def cancel_number(self, params=None, **kwargs):
        return self._client.post(self._client.host(), "/number/cancel", params or kwargs, **Numbers.defaults)

    def update_number(self, params=None, **kwargs):
        return self._client.post(self._client.host(), "/number/update", params or kwargs, **Numbers.defaults)



from __future__ import annotations
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from vonage import Client

from pydantic import BaseModel, ValidationError, validator, conint, constr
from typing import Optional, List

import copy
import re

from ._internal import set_auth_type
from .errors import Verify2Error


class Verify2:
    valid_channels = [
        'sms',
        'whatsapp',
        'whatsapp_interactive',
        'voice',
        'email',
        'silent_auth',
    ]

    def __init__(self, client: Client):
        self._client = client
        self._auth_type = set_auth_type(self._client)

    def new_request(self, params: dict):
        self._remove_unnecessary_fraud_check(params)
        try:
            params_to_verify = copy.deepcopy(params)
            Verify2.VerifyRequest.parse_obj(params_to_verify)
        except (ValidationError, Verify2Error) as err:
            raise err

        return self._client.post(
            self._client.api_host(),
            '/v2/verify',
            params,
            auth_type=self._auth_type,
        )

    def check_code(self, request_id: str, code: str):
        params = {'code': str(code)}

        return self._client.post(
            self._client.api_host(),
            f'/v2/verify/{request_id}',
            params,
            auth_type=self._auth_type,
        )

    def cancel_verification(self, request_id: str):
        return self._client.delete(
            self._client.api_host(),
            f'/v2/verify/{request_id}',
            auth_type=self._auth_type,
        )

    def _remove_unnecessary_fraud_check(self, params):
        if 'fraud_check' in params and params['fraud_check'] != False:
            del params['fraud_check']

    class VerifyRequest(BaseModel):
        brand: str
        workflow: List[dict]
        locale: Optional[str]
        channel_timeout: Optional[conint(ge=60, le=900)]
        client_ref: Optional[str]
        code_length: Optional[conint(ge=4, le=10)]
        fraud_check: Optional[bool]
        code: Optional[
            constr(min_length=4, max_length=10, regex='^(?=[a-zA-Z0-9]{4,10}$)[a-zA-Z0-9]*$')
        ]

        @validator('workflow')
        def check_valid_workflow(cls, v):
            for workflow in v:
                Verify2._check_valid_channel(workflow)
                Verify2._check_valid_recipient(workflow)
                Verify2._check_app_hash(workflow)
                if workflow['channel'] == 'whatsapp' and 'from' in workflow:
                    Verify2._check_whatsapp_sender(workflow)

    def _check_valid_channel(workflow):
        if 'channel' not in workflow or workflow['channel'] not in Verify2.valid_channels:
            raise Verify2Error(
                f'You must specify a valid verify channel inside the "workflow" object, one of: "{Verify2.valid_channels}"'
            )

    def _check_valid_recipient(workflow):
        if 'to' not in workflow or (
            workflow['channel'] != 'email' and not re.search(r'^[1-9]\d{6,14}$', workflow['to'])
        ):
            raise Verify2Error(
                f'You must specify a valid "to" value for channel "{workflow["channel"]}"'
            )

    def _check_app_hash(workflow):
        if workflow['channel'] == 'sms' and 'app_hash' in workflow:
            if type(workflow['app_hash']) != str or len(workflow['app_hash']) != 11:
                raise Verify2Error(
                    'Invalid "app_hash" specified. If specifying app_hash, \
                        it must be passed as a string and contain exactly 11 characters.'
                )
        elif workflow['channel'] != 'sms' and 'app_hash' in workflow:
            raise Verify2Error(
                'Cannot specify a value for "app_hash" unless using SMS for authentication.'
            )

    def _check_whatsapp_sender(workflow):
        if not re.search(r'^[1-9]\d{6,14}$', workflow['from']):
            raise Verify2Error(f'You must specify a valid "from" value if included.')



class Ussd:
    defaults = {'auth_type': 'params', 'body_is_json': False}
    
    def __init__(self, client):
        self._client = client

    def send_ussd_push_message(self, params=None, **kwargs):
        return self._client.post(self._client.host(), "/ussd/json", params or kwargs, **Ussd.defaults)

    def send_ussd_prompt_message(self, params=None, **kwargs):
        return self._client.post(self._client.host(), "/ussd-prompt/json", params or kwargs, **Ussd.defaults)



from .errors import MeetingsError

from typing_extensions import Literal
import logging
import requests


logger = logging.getLogger("vonage")


class Meetings:
    """Class containing methods used to create and manage meetings using the Meetings API."""

    _auth_type = 'jwt'

    def __init__(self, client):
        self._client = client
        self._meetings_api_host = client.meetings_api_host()

    def list_rooms(self, page_size: str = 20, start_id: str = None, end_id: str = None):
        params = Meetings.set_start_and_end_params(start_id, end_id)
        params['page_size'] = page_size
        return self._client.get(
            self._meetings_api_host, '/rooms', params, auth_type=Meetings._auth_type
        )

    def create_room(self, params: dict = {}):
        if 'display_name' not in params:
            raise MeetingsError(
                'You must include a value for display_name as a field in the params dict when creating a meeting room.'
            )
        if 'type' not in params or 'type' in params and params['type'] != 'long_term':
            if 'expires_at' in params:
                raise MeetingsError('Cannot set "expires_at" for an instant room.')
        elif params['type'] == 'long_term' and 'expires_at' not in params:
            raise MeetingsError('You must set a value for "expires_at" for a long-term room.')

        return self._client.post(
            self._meetings_api_host, '/rooms', params, auth_type=Meetings._auth_type
        )

    def get_room(self, room_id: str):
        return self._client.get(
            self._meetings_api_host, f'/rooms/{room_id}', auth_type=Meetings._auth_type
        )

    def update_room(self, room_id: str, params: dict):
        return self._client.patch(
            self._meetings_api_host, f'/rooms/{room_id}', params, auth_type=Meetings._auth_type
        )

    def add_theme_to_room(self, room_id: str, theme_id: str):
        params = {'update_details': {'theme_id': theme_id}}
        return self._client.patch(
            self._meetings_api_host, f'/rooms/{room_id}', params, auth_type=Meetings._auth_type
        )

    def get_recording(self, recording_id: str):
        return self._client.get(
            self._meetings_api_host, f'/recordings/{recording_id}', auth_type=Meetings._auth_type
        )

    def delete_recording(self, recording_id: str):
        return self._client.delete(
            self._meetings_api_host, f'/recordings/{recording_id}', auth_type=Meetings._auth_type
        )

    def get_session_recordings(self, session_id: str):
        return self._client.get(
            self._meetings_api_host,
            f'/sessions/{session_id}/recordings',
            auth_type=Meetings._auth_type,
        )

    def list_dial_in_numbers(self):
        return self._client.get(
            self._meetings_api_host, '/dial-in-numbers', auth_type=Meetings._auth_type
        )

    def list_themes(self):
        return self._client.get(self._meetings_api_host, '/themes', auth_type=Meetings._auth_type)

    def create_theme(self, params: dict):
        if 'main_color' not in params or 'brand_text' not in params:
            raise MeetingsError('Values for "main_color" and "brand_text" must be specified')

        return self._client.post(
            self._meetings_api_host, '/themes', params, auth_type=Meetings._auth_type
        )

    def get_theme(self, theme_id: str):
        return self._client.get(
            self._meetings_api_host, f'/themes/{theme_id}', auth_type=Meetings._auth_type
        )

    def delete_theme(self, theme_id: str, force: bool = False):
        params = {'force': force}
        return self._client.delete(
            self._meetings_api_host,
            f'/themes/{theme_id}',
            params=params,
            auth_type=Meetings._auth_type,
        )

    def update_theme(self, theme_id: str, params: dict):
        return self._client.patch(
            self._meetings_api_host, f'/themes/{theme_id}', params, auth_type=Meetings._auth_type
        )

    def list_rooms_with_theme_id(
        self, theme_id: str, page_size: int = 20, start_id: str = None, end_id: str = None
    ):
        params = Meetings.set_start_and_end_params(start_id, end_id)
        params['page_size'] = page_size

        return self._client.get(
            self._meetings_api_host,
            f'/themes/{theme_id}/rooms',
            params,
            auth_type=Meetings._auth_type,
        )

    def update_application_theme(self, theme_id: str):
        params = {'update_details': {'default_theme_id': theme_id}}
        return self._client.patch(
            self._meetings_api_host, '/applications', params, auth_type=Meetings._auth_type
        )

    def upload_logo_to_theme(
        self, theme_id: str, path_to_image: str, logo_type: Literal['white', 'colored', 'favicon']
    ):
        params = self._get_logo_upload_url(logo_type)
        self._upload_to_aws(params, path_to_image)
        self._add_logo_to_theme(theme_id, params['fields']['key'])
        return f'Logo upload to theme: {theme_id} was successful.'

    def _get_logo_upload_url(self, logo_type):
        upload_urls = self._client.get(
            self._meetings_api_host, '/themes/logos-upload-urls', auth_type=Meetings._auth_type
        )
        for url_object in upload_urls:
            if url_object['fields']['logoType'] == logo_type:
                return url_object
        raise MeetingsError('Cannot find the upload URL for the specified logo type.')

    def _upload_to_aws(self, params, path_to_image):
        form = {**params['fields'], 'file': open(path_to_image, 'rb')}

        logger.debug(f"POST to {params['url']} to upload file {path_to_image}")
        logo_upload = requests.post(
            url=params['url'],
            files=form,
        )
        if logo_upload.status_code != 204:
            raise MeetingsError(f'Logo upload process failed. {logo_upload.content}')

    def _add_logo_to_theme(self, theme_id: str, key: str):
        params = {'keys': [key]}
        return self._client.put(
            self._meetings_api_host,
            f'/themes/{theme_id}/finalizeLogos',
            params,
            auth_type=Meetings._auth_type,
        )

    @staticmethod
    def set_start_and_end_params(start_id, end_id):
        params = {}
        if start_id is not None:
            params['start_id'] = start_id
        if end_id is not None:
            params['end_id'] = end_id
        return params



def test_getters(client, dummy_data):
    assert client.host() == dummy_data.host
    assert client.api_host() == dummy_data.api_host

def test_setters(client, dummy_data):
    try:
        client.host('host.vonage.com')
        client.api_host('host.vonage.com')
        assert client.host() != dummy_data.host
        assert client.api_host() != dummy_data.api_host
    except:
        assert False



from pydantic import BaseModel, confloat, conint
from typing import Optional, List


class InputTypes:
    class Dtmf(BaseModel):
        timeOut: Optional[conint(ge=0, le=10)]
        maxDigits: Optional[conint(ge=1, le=20)]
        submitOnHash: Optional[bool]

    class Speech(BaseModel):
        uuid: Optional[str]
        endOnSilence: Optional[confloat(ge=0.4, le=10.0)]
        language: Optional[str]
        context: Optional[List[str]]
        startTimeout: Optional[conint(ge=1, le=60)]
        maxDuration: Optional[conint(ge=1, le=60)]
        saveAudio: Optional[bool]

    @classmethod
    def create_dtmf_model(cls, dict) -> Dtmf:
        return cls.Dtmf.parse_obj(dict)

    @classmethod
    def create_speech_model(cls, dict) -> Speech:
        return cls.Speech.parse_obj(dict)



from __future__ import annotations
from typing import TYPE_CHECKING, Optional, Union

from .errors import SubaccountsError

if TYPE_CHECKING:
    from vonage import Client


class Subaccounts:
    """Class containing methods for working with the Vonage Subaccounts API."""

    default_start_date = '1970-01-01T00:00:00Z'

    def __init__(self, client: Client):
        self._client = client
        self._api_key = self._client.api_key
        self._api_host = self._client.api_host()
        self._auth_type = 'header'

    def list_subaccounts(self):
        return self._client.get(
            self._api_host,
            f'/accounts/{self._api_key}/subaccounts',
            auth_type=self._auth_type,
        )

    def create_subaccount(
        self,
        name: str,
        secret: Optional[str] = None,
        use_primary_account_balance: Optional[bool] = None,
    ):
        params = {'name': name, 'secret': secret}
        if self._is_boolean(use_primary_account_balance):
            params['use_primary_account_balance'] = use_primary_account_balance

        return self._client.post(
            self._api_host,
            f'/accounts/{self._api_key}/subaccounts',
            params=params,
            auth_type=self._auth_type,
        )

    def get_subaccount(self, subaccount_key: str):
        return self._client.get(
            self._api_host,
            f'/accounts/{self._api_key}/subaccounts/{subaccount_key}',
            auth_type=self._auth_type,
        )

    def modify_subaccount(
        self,
        subaccount_key: str,
        suspended: Optional[bool] = None,
        use_primary_account_balance: Optional[bool] = None,
        name: Optional[str] = None,
    ):
        params = {'name': name}
        if self._is_boolean(suspended):
            params['suspended'] = suspended
        if self._is_boolean(use_primary_account_balance):
            params['use_primary_account_balance'] = use_primary_account_balance

        return self._client.patch(
            self._api_host,
            f'/accounts/{self._api_key}/subaccounts/{subaccount_key}',
            params=params,
            auth_type=self._auth_type,
        )

    def list_credit_transfers(
        self,
        start_date: str = default_start_date,
        end_date: Optional[str] = None,
        subaccount: Optional[str] = None,
    ):
        params = {
            'start_date': start_date,
            'end_date': end_date,
            'subaccount': subaccount,
        }

        return self._client.get(
            self._api_host,
            f'/accounts/{self._api_key}/credit-transfers',
            params=params,
            auth_type=self._auth_type,
        )

    def transfer_credit(
        self,
        from_: str,
        to: str,
        amount: Union[float, int],
        reference: str = None,
    ):
        params = {
            'from': from_,
            'to': to,
            'amount': amount,
            'reference': reference,
        }

        return self._client.post(
            self._api_host,
            f'/accounts/{self._api_key}/credit-transfers',
            params=params,
            auth_type=self._auth_type,
        )

    def list_balance_transfers(
        self,
        start_date: str = default_start_date,
        end_date: Optional[str] = None,
        subaccount: Optional[str] = None,
    ):
        params = {
            'start_date': start_date,
            'end_date': end_date,
            'subaccount': subaccount,
        }

        return self._client.get(
            self._api_host,
            f'/accounts/{self._api_key}/balance-transfers',
            params=params,
            auth_type=self._auth_type,
        )

    def transfer_balance(
        self,
        from_: str,
        to: str,
        amount: Union[float, int],
        reference: str = None,
    ):
        params = {'from': from_, 'to': to, 'amount': amount, 'reference': reference}

        return self._client.post(
            self._api_host,
            f'/accounts/{self._api_key}/balance-transfers',
            params=params,
            auth_type=self._auth_type,
        )

    def transfer_number(self, from_: str, to: str, number: int, country: str):
        params = {'from': from_, 'to': to, 'number': number, 'country': country}
        return self._client.post(
            self._api_host,
            f'/accounts/{self._api_key}/transfer-number',
            params=params,
            auth_type=self._auth_type,
        )

    def _is_boolean(self, var):
        if var is not None:
            if type(var) == bool:
                return True
            else:
                raise SubaccountsError(
                    f'If providing a value, it needs to be a boolean. You provided: "{var}"'
                )



import os.path
import time

import jwt

import vonage
from vonage import Ncco
from util import *


@responses.activate
def test_create_call(voice, dummy_data):
    stub(responses.POST, "https://api.nexmo.com/v1/calls")

    params = {
        "to": [{"type": "phone", "number": "14843331234"}],
        "from": {"type": "phone", "number": "14843335555"},
        "answer_url": ["https://example.com/answer"],
    }

    assert isinstance(voice.create_call(params), dict)
    assert request_user_agent() == dummy_data.user_agent
    assert request_content_type() == "application/json"


@responses.activate
def test_params_with_random_number(voice, dummy_data):
    stub(responses.POST, "https://api.nexmo.com/v1/calls")

    params = {
        "to": [{"type": "phone", "number": "14843331234"}],
        "random_from_number": True,
        "answer_url": ["https://example.com/answer"],
    }

    assert isinstance(voice.create_call(params), dict)
    assert request_user_agent() == dummy_data.user_agent
    assert request_content_type() == "application/json"


@responses.activate
def test_create_call_with_ncco_builder(voice, dummy_data):
    stub(responses.POST, "https://api.nexmo.com/v1/calls")

    talk = Ncco.Talk(
        text='Hello from Vonage!', bargeIn=True, loop=3, level=0.5, language='en-GB', style=1, premium=True
    )
    ncco = Ncco.build_ncco(talk)
    voice.create_call(
        {
            'to': [{'type': 'phone', 'number': '447449815316'}],
            'from': {'type': 'phone', 'number': '447418370240'},
            'ncco': ncco,
        }
    )
    assert (
        request_body()
        == b'{"to": [{"type": "phone", "number": "447449815316"}], "from": {"type": "phone", "number": "447418370240"}, "ncco": [{"action": "talk", "text": "Hello from Vonage!", "bargeIn": true, "loop": 3, "level": 0.5, "language": "en-GB", "style": 1, "premium": true}]}'
    )
    assert request_user_agent() == dummy_data.user_agent
    assert request_content_type() == "application/json"


@responses.activate
def test_get_calls(voice, dummy_data):
    stub(responses.GET, "https://api.nexmo.com/v1/calls")

    assert isinstance(voice.get_calls(), dict)
    assert request_user_agent() == dummy_data.user_agent
    assert_re(r"\ABearer ", request_authorization())


@responses.activate
def test_get_call(voice, dummy_data):
    stub(responses.GET, "https://api.nexmo.com/v1/calls/xx-xx-xx-xx")

    assert isinstance(voice.get_call("xx-xx-xx-xx"), dict)
    assert request_user_agent() == dummy_data.user_agent
    assert_re(r"\ABearer ", request_authorization())


@responses.activate
def test_update_call(voice, dummy_data):
    stub(responses.PUT, "https://api.nexmo.com/v1/calls/xx-xx-xx-xx")

    assert isinstance(voice.update_call("xx-xx-xx-xx", action="hangup"), dict)
    assert request_user_agent() == dummy_data.user_agent
    assert request_content_type() == "application/json"
    assert request_body() == b'{"action": "hangup"}'


@responses.activate
def test_send_audio(voice, dummy_data):
    stub(responses.PUT, "https://api.nexmo.com/v1/calls/xx-xx-xx-xx/stream")

    assert isinstance(
        voice.send_audio("xx-xx-xx-xx", stream_url="http://example.com/audio.mp3"),
        dict,
    )
    assert request_user_agent() == dummy_data.user_agent
    assert request_content_type() == "application/json"
    assert request_body() == b'{"stream_url": "http://example.com/audio.mp3"}'


@responses.activate
def test_stop_audio(voice, dummy_data):
    stub(responses.DELETE, "https://api.nexmo.com/v1/calls/xx-xx-xx-xx/stream")

    assert isinstance(voice.stop_audio("xx-xx-xx-xx"), dict)
    assert request_user_agent() == dummy_data.user_agent


@responses.activate
def test_send_speech(voice, dummy_data):
    stub(responses.PUT, "https://api.nexmo.com/v1/calls/xx-xx-xx-xx/talk")

    assert isinstance(voice.send_speech("xx-xx-xx-xx", text="Hello"), dict)
    assert request_user_agent() == dummy_data.user_agent
    assert request_content_type() == "application/json"
    assert request_body() == b'{"text": "Hello"}'


@responses.activate
def test_stop_speech(voice, dummy_data):
    stub(responses.DELETE, "https://api.nexmo.com/v1/calls/xx-xx-xx-xx/talk")

    assert isinstance(voice.stop_speech("xx-xx-xx-xx"), dict)
    assert request_user_agent() == dummy_data.user_agent


@responses.activate
def test_send_dtmf(voice, dummy_data):
    stub(responses.PUT, "https://api.nexmo.com/v1/calls/xx-xx-xx-xx/dtmf")

    assert isinstance(voice.send_dtmf("xx-xx-xx-xx", digits="1234"), dict)
    assert request_user_agent() == dummy_data.user_agent
    assert request_content_type() == "application/json"
    assert request_body() == b'{"digits": "1234"}'


@responses.activate
def test_user_provided_authorization(dummy_data):
    stub(responses.GET, "https://api.nexmo.com/v1/calls/xx-xx-xx-xx")

    application_id = "different-application-id"
    client = vonage.Client(application_id=application_id, private_key=dummy_data.private_key)

    nbf = int(time.time())
    exp = nbf + 3600
    
    client.auth(nbf=nbf, exp=exp)
    client.voice.get_call("xx-xx-xx-xx")

    token = request_authorization().split()[1]

    token = jwt.decode(token, dummy_data.public_key, algorithms="RS256")
    print(token)
    assert token["application_id"] == application_id
    assert token["nbf"] == nbf
    assert token["exp"] == exp


@responses.activate
def test_authorization_with_private_key_path(dummy_data):
    stub(responses.GET, "https://api.nexmo.com/v1/calls/xx-xx-xx-xx")

    private_key = os.path.join(os.path.dirname(__file__), "data/private_key.txt")

    client = vonage.Client(
        key=dummy_data.api_key,
        secret=dummy_data.api_secret,
        application_id=dummy_data.application_id,
        private_key=private_key,
    )
    voice = vonage.Voice(client)
    voice.get_call("xx-xx-xx-xx")

    token = jwt.decode(request_authorization().split()[1], dummy_data.public_key, algorithms="RS256")
    assert token["application_id"] == dummy_data.application_id


@responses.activate
def test_authorization_with_private_key_object(voice, dummy_data):
    stub(responses.GET, "https://api.nexmo.com/v1/calls/xx-xx-xx-xx")

    voice.get_call("xx-xx-xx-xx")

    token = jwt.decode(request_authorization().split()[1], dummy_data.public_key, algorithms="RS256")
    assert token["application_id"] == dummy_data.application_id


@responses.activate
def test_get_recording(voice, dummy_data):
    stub_bytes(
        responses.GET, "https://api.nexmo.com/v1/files/d6e47a2e-3414-11e8-8c2c-2f8b643ed957", body=b'THISISANMP3'
    )

    assert isinstance(
        voice.get_recording("https://api.nexmo.com/v1/files/d6e47a2e-3414-11e8-8c2c-2f8b643ed957"),
        bytes,
    )
    assert request_user_agent() == dummy_data.user_agent



from util import *
from vonage.errors import RedactError


def test_redact_invalid_product_name(redact):
    with pytest.raises(RedactError):
        redact.redact_transaction(id='not-a-real-id', product='fake-product')

@responses.activate
def test_redact_transaction(redact, dummy_data):
    responses.add(
        responses.POST,
        "https://api.nexmo.com/v1/redact/transaction",
        body=None,
        status=204,
    )

    assert redact.redact_transaction(id="not-a-real-id", product="sms") is None
    assert request_user_agent() == dummy_data.user_agent
    assert request_content_type() == "application/json"


@responses.activate
def test_redact_transaction_with_type(redact, dummy_data):
    responses.add(
        responses.POST,
        "https://api.nexmo.com/v1/redact/transaction",
        body=None,
        status=204,
    )

    assert redact.redact_transaction(id="some-id", product="sms", type="xyz") is None
    assert request_user_agent() == dummy_data.user_agent
    assert request_content_type() == "application/json"
    assert b"xyz" in request_body()



import json
from util import *

import vonage

@responses.activate
def test_deprecated_list_applications(application_v2, dummy_data):
    stub(
        responses.GET,
        "https://api.nexmo.com/v2/applications",
        fixture_path="applications/list_applications.json",
    )

    apps = application_v2.list_applications()
    assert_basic_auth()
    assert isinstance(apps, dict)
    assert apps["total_items"] == 30
    assert request_user_agent() == dummy_data.user_agent


@responses.activate
def test_deprecated_get_application(application_v2, dummy_data):
    stub(
        responses.GET,
        "https://api.nexmo.com/v2/applications/xx-xx-xx-xx",
        fixture_path="applications/get_application.json",
    )

    app = application_v2.get_application("xx-xx-xx-xx")
    assert_basic_auth()
    assert isinstance(app, dict)
    assert app["name"] == "My Test Application"
    assert request_user_agent() == dummy_data.user_agent


@responses.activate
def test_deprecated_create_application(application_v2, dummy_data):
    stub(
        responses.POST,
        "https://api.nexmo.com/v2/applications",
        fixture_path="applications/create_application.json",
    )

    params = {"name": "Example App", "type": "voice"}

    app = application_v2.create_application(params)
    assert_basic_auth()
    assert isinstance(app, dict)
    assert app["name"] == "My Test Application"
    assert request_user_agent() == dummy_data.user_agent
    body_data = json.loads(request_body().decode("utf-8"))
    assert body_data["type"] == "voice"


@responses.activate
def test_deprecated_update_application(application_v2, dummy_data):
    stub(
        responses.PUT,
        "https://api.nexmo.com/v2/applications/xx-xx-xx-xx",
        fixture_path="applications/update_application.json",
    )

    params = {"answer_url": "https://example.com/ncco"}

    app = application_v2.update_application("xx-xx-xx-xx", params)
    assert_basic_auth()
    assert isinstance(app, dict)
    assert request_user_agent() == dummy_data.user_agent
    assert request_content_type() == "application/json"
    assert b'"answer_url": "https://example.com/ncco"' in request_body()

    assert app["name"] == "A Better Name"


@responses.activate
def test_deprecated_delete_application(application_v2, dummy_data):
    responses.add(
        responses.DELETE,
        "https://api.nexmo.com/v2/applications/xx-xx-xx-xx",
        status=204,
    )

    assert application_v2.delete_application("xx-xx-xx-xx") is None
    assert_basic_auth()
    assert request_user_agent() == dummy_data.user_agent


@responses.activate
def test_deprecated_authentication_error(application_v2):
    responses.add(
        responses.DELETE,
        "https://api.nexmo.com/v2/applications/xx-xx-xx-xx",
        status=401,
    )
    with pytest.raises(vonage.AuthenticationError):
        application_v2.delete_application("xx-xx-xx-xx")


@responses.activate
def test_deprecated_client_error(application_v2):
    responses.add(
        responses.DELETE,
        "https://api.nexmo.com/v2/applications/xx-xx-xx-xx",
        status=430,
        body=json.dumps(
            {
                "type": "nope_error",
                "title": "Nope",
                "detail": "You really shouldn't have done that",
            }
        ),
    )
    with pytest.raises(vonage.ClientError) as exc_info:
        application_v2.delete_application("xx-xx-xx-xx")
    assert (
        str(exc_info.value) == "Nope: You really shouldn't have done that (nope_error)"
    )


@responses.activate
def test_deprecated_client_error_no_decode(application_v2):
    responses.add(
        responses.DELETE,
        "https://api.nexmo.com/v2/applications/xx-xx-xx-xx",
        status=430,
        body="{this: isnot_json",
    )
    with pytest.raises(vonage.ClientError) as exc_info:
        application_v2.delete_application("xx-xx-xx-xx")
    assert str(exc_info.value) == "430 response from api.nexmo.com"


@responses.activate
def test_deprecated_server_error(application_v2):
    responses.add(
        responses.DELETE,
        "https://api.nexmo.com/v2/applications/xx-xx-xx-xx",
        status=500,
    )
    with pytest.raises(vonage.ServerError):
        application_v2.delete_application("xx-xx-xx-xx")



@responses.activate
def test_list_applications(client, dummy_data):
    stub(
        responses.GET,
        "https://api.nexmo.com/v2/applications",
        fixture_path="applications/list_applications.json",
    )

    apps = client.application.list_applications()
    assert_basic_auth()
    assert isinstance(apps, dict)
    assert apps["total_items"] == 30
    assert request_user_agent() == dummy_data.user_agent


@responses.activate
def test_get_application(client, dummy_data):
    stub(
        responses.GET,
        "https://api.nexmo.com/v2/applications/xx-xx-xx-xx",
        fixture_path="applications/get_application.json",
    )

    app = client.application.get_application("xx-xx-xx-xx")
    assert_basic_auth()
    assert isinstance(app, dict)
    assert app["name"] == "My Test Application"
    assert request_user_agent() == dummy_data.user_agent


@responses.activate
def test_create_application(client, dummy_data):
    stub(
        responses.POST,
        "https://api.nexmo.com/v2/applications",
        fixture_path="applications/create_application.json",
    )

    params = {"name": "Example App", "type": "voice"}

    app = client.application.create_application(params)
    assert_basic_auth()
    assert isinstance(app, dict)
    assert app["name"] == "My Test Application"
    assert request_user_agent() == dummy_data.user_agent
    body_data = json.loads(request_body().decode("utf-8"))
    assert body_data["type"] == "voice"


@responses.activate
def test_update_application(client, dummy_data):
    stub(
        responses.PUT,
        "https://api.nexmo.com/v2/applications/xx-xx-xx-xx",
        fixture_path="applications/update_application.json",
    )

    params = {"answer_url": "https://example.com/ncco"}

    app = client.application.update_application("xx-xx-xx-xx", params)
    assert_basic_auth()
    assert isinstance(app, dict)
    assert request_user_agent() == dummy_data.user_agent
    assert request_content_type() == "application/json"
    assert b'"answer_url": "https://example.com/ncco"' in request_body()

    assert app["name"] == "A Better Name"


@responses.activate
def test_delete_application(client, dummy_data):
    responses.add(
        responses.DELETE,
        "https://api.nexmo.com/v2/applications/xx-xx-xx-xx",
        status=204,
    )

    assert client.application.delete_application("xx-xx-xx-xx") is None
    assert_basic_auth()
    assert request_user_agent() == dummy_data.user_agent


@responses.activate
def test_authentication_error(client):
    responses.add(
        responses.DELETE,
        "https://api.nexmo.com/v2/applications/xx-xx-xx-xx",
        status=401,
    )
    with pytest.raises(vonage.AuthenticationError):
        client.application.delete_application("xx-xx-xx-xx")


@responses.activate
def test_client_error(client):
    responses.add(
        responses.DELETE,
        "https://api.nexmo.com/v2/applications/xx-xx-xx-xx",
        status=430,
        body=json.dumps(
            {
                "type": "nope_error",
                "title": "Nope",
                "detail": "You really shouldn't have done that",
            }
        ),
    )
    with pytest.raises(vonage.ClientError) as exc_info:
        client.application.delete_application("xx-xx-xx-xx")
    assert (
        str(exc_info.value) == "Nope: You really shouldn't have done that (nope_error)"
    )


@responses.activate
def test_client_error_no_decode(client):
    responses.add(
        responses.DELETE,
        "https://api.nexmo.com/v2/applications/xx-xx-xx-xx",
        status=430,
        body="{this: isnot_json",
    )
    with pytest.raises(vonage.ClientError) as exc_info:
        client.application.delete_application("xx-xx-xx-xx")
    assert str(exc_info.value) == "430 response from api.nexmo.com"


@responses.activate
def test_server_error(client):
    responses.add(
        responses.DELETE,
        "https://api.nexmo.com/v2/applications/xx-xx-xx-xx",
        status=500,
    )
    with pytest.raises(vonage.ServerError):
        client.application.delete_application("xx-xx-xx-xx")



import vonage
from util import *
from vonage.errors import InvalidAuthenticationTypeError


def test_client_doesnt_require_api_key(dummy_data):
    client = vonage.Client(application_id="myid", private_key=dummy_data.private_key)
    assert client is not None
    assert client.api_key is None
    assert client.api_secret is None


@responses.activate
def test_client_can_make_application_requests_without_api_key(dummy_data):
    stub(responses.POST, "https://api.nexmo.com/v1/calls")

    client = vonage.Client(application_id="myid", private_key=dummy_data.private_key)
    voice = vonage.Voice(client)
    voice.create_call("123455")


def test_invalid_auth_type_raises_error(client):
    with pytest.raises(InvalidAuthenticationTypeError):
        client.get(client.host(), 'my/request/uri', auth_type='magic')

@responses.activate
def test_timeout_is_set_on_client_calls(dummy_data):
    stub(responses.POST, "https://api.nexmo.com/v1/calls")

    client = vonage.Client(application_id="myid", private_key=dummy_data.private_key, timeout=1)
    voice = vonage.Voice(client)
    voice.create_call("123455")

    assert len(responses.calls) == 1
    assert responses.calls[0].request.req_kwargs["timeout"] == 1



from .errors import CallbackRequiredError

class NumberInsight:
    auth_type = 'params'
    
    def __init__(self, client):
        self._client = client

    def get_basic_number_insight(self, params=None, **kwargs):
        return self._client.get(self._client.api_host(), "/ni/basic/json", params or kwargs, auth_type=NumberInsight.auth_type)

    def get_standard_number_insight(self, params=None, **kwargs):
        return self._client.get(self._client.api_host(), "/ni/standard/json", params or kwargs, auth_type=NumberInsight.auth_type)

    def get_advanced_number_insight(self, params=None, **kwargs):
        return self._client.get(self._client.api_host(), "/ni/advanced/json", params or kwargs, auth_type=NumberInsight.auth_type)

    def get_async_advanced_number_insight(self, params=None, **kwargs):
        argoparams = params or kwargs
        if "callback" in argoparams and type(argoparams["callback"]) == str and argoparams["callback"] != "":
            return self._client.get(
                self._client.api_host(), "/ni/advanced/async/json", params or kwargs, auth_type=NumberInsight.auth_type
            )
        else:
            raise CallbackRequiredError(
                "A callback is needed for async advanced number insight"
            )
    


from util import *


@responses.activate
def test_get_account_numbers(numbers, dummy_data):
    stub(responses.GET, "https://rest.nexmo.com/account/numbers")

    assert isinstance(numbers.get_account_numbers(size=25), dict)
    assert request_user_agent() == dummy_data.user_agent
    assert request_params()["size"] == ["25"]


@responses.activate
def test_get_available_numbers(numbers, dummy_data):
    stub(responses.GET, "https://rest.nexmo.com/number/search")

    assert isinstance(numbers.get_available_numbers("CA", size=25), dict)
    assert request_user_agent() == dummy_data.user_agent
    assert "country=CA" in request_query()
    assert "size=25" in request_query()


@responses.activate
def test_buy_number(numbers, dummy_data):
    stub(responses.POST, "https://rest.nexmo.com/number/buy")

    params = {"country": "US", "msisdn": "number"}

    assert isinstance(numbers.buy_number(params), dict)
    assert request_user_agent() == dummy_data.user_agent
    assert "country=US" in request_body()
    assert "msisdn=number" in request_body()


@responses.activate
def test_cancel_number(numbers, dummy_data):
    stub(responses.POST, "https://rest.nexmo.com/number/cancel")

    params = {"country": "US", "msisdn": "number"}

    assert isinstance(numbers.cancel_number(params), dict)
    assert request_user_agent() == dummy_data.user_agent
    assert "country=US" in request_body()
    assert "msisdn=number" in request_body()


@responses.activate
def test_update_number(numbers, dummy_data):
    stub(responses.POST, "https://rest.nexmo.com/number/update")

    params = {"country": "US", "msisdn": "number", "moHttpUrl": "callback"}

    assert isinstance(numbers.update_number(params), dict)
    assert request_user_agent() == dummy_data.user_agent
    assert "country=US" in request_body()
    assert "msisdn=number" in request_body()
    assert "moHttpUrl=callback" in request_body()



from __future__ import annotations
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from vonage import Client

from .errors import UsersError
from ._internal import set_auth_type


class Users:
    """Class containing methods for user management as part of the Application API."""

    def __init__(self, client: Client):
        self._client = client
        self._auth_type = set_auth_type(self._client)

    def list_users(
        self,
        page_size: int = None,
        order: str = 'asc',
        cursor: str = None,
        name: str = None,
    ):
        """
        Lists the name and user id of all users associated with the account.
        For complete information on a user, call Users.get_user, passing in the user id.
        """

        if order.lower() not in ('asc', 'desc'):
            raise UsersError(
                'Invalid order parameter. Must be one of: "asc", "desc", "ASC", "DESC".'
            )

        params = {'page_size': page_size, 'order': order.lower(), 'cursor': cursor, 'name': name}
        return self._client.get(
            self._client.api_host(),
            '/v1/users',
            params,
            auth_type=self._auth_type,
        )

    def create_user(self, params: dict = None):
        self._client.headers['Content-Type'] = 'application/json'
        return self._client.post(
            self._client.api_host(),
            '/v1/users',
            params,
            auth_type=self._auth_type,
        )

    def get_user(self, user_id: str):
        return self._client.get(
            self._client.api_host(),
            f'/v1/users/{user_id}',
            auth_type=self._auth_type,
        )

    def update_user(self, user_id: str, params: dict):
        return self._client.patch(
            self._client.api_host(),
            f'/v1/users/{user_id}',
            params,
            auth_type=self._auth_type,
        )

    def delete_user(self, user_id: str):
        return self._client.delete(
            self._client.api_host(),
            f'/v1/users/{user_id}',
            auth_type=self._auth_type,
        )



from util import *

@responses.activate
def test_start_verification(verify, dummy_data):
    stub(responses.POST, "https://api.nexmo.com/verify/json")

    params = {"number": "447525856424", "brand": "MyApp"}

    assert isinstance(verify.start_verification(params), dict)
    assert request_user_agent() == dummy_data.user_agent
    assert "number=447525856424" in request_body()
    assert "brand=MyApp" in request_body()


@responses.activate
def test_check_verification(verify, dummy_data):
    stub(responses.POST, "https://api.nexmo.com/verify/check/json")

    assert isinstance(verify.check("8g88g88eg8g8gg9g90", code="123445"), dict)
    assert request_user_agent() == dummy_data.user_agent
    assert "code=123445" in request_body()
    assert "request_id=8g88g88eg8g8gg9g90" in request_body()


@responses.activate
def test_get_verification(verify, dummy_data):
    stub(responses.GET, "https://api.nexmo.com/verify/search/json")

    assert isinstance(verify.search("xxx"), dict)
    assert request_user_agent() == dummy_data.user_agent
    assert "request_id=xxx" in request_query()


@responses.activate
def test_cancel_verification(verify, dummy_data):
    stub(responses.POST, "https://api.nexmo.com/verify/control/json")

    assert isinstance(verify.cancel("8g88g88eg8g8gg9g90"), dict)
    assert request_user_agent() == dummy_data.user_agent
    assert "cmd=cancel" in request_body()
    assert "request_id=8g88g88eg8g8gg9g90" in request_body()


@responses.activate
def test_trigger_next_verification_event(verify, dummy_data):
    stub(responses.POST, "https://api.nexmo.com/verify/control/json")

    assert isinstance(verify.trigger_next_event("8g88g88eg8g8gg9g90"), dict)
    assert request_user_agent() == dummy_data.user_agent
    assert "cmd=trigger_next_event" in request_body()
    assert "request_id=8g88g88eg8g8gg9g90" in request_body()


@responses.activate
def test_start_psd2_verification(verify, dummy_data):
    stub(responses.POST, "https://api.nexmo.com/verify/psd2/json")

    params = {"number": "447525856424", "brand": "MyApp"}

    assert isinstance(verify.psd2(params), dict)
    assert request_user_agent() == dummy_data.user_agent
    assert "number=447525856424" in request_body()
    assert "brand=MyApp" in request_body()


@responses.activate
def test_start_verification_blacklisted_error_with_network(client, dummy_data):
    stub(responses.POST, "https://api.nexmo.com/verify/json",
        fixture_path="verify/blocked_with_network.json"
    )

    params = {"number": "447525856424", "brand": "MyApp"}
    response = client.verify.start_verification(params)

    assert isinstance(response, dict)
    assert request_user_agent() == dummy_data.user_agent
    assert "number=447525856424" in request_body()
    assert "brand=MyApp" in request_body()
    assert response["status"] == "7"
    assert response["network"] == "25503"
    assert response["error_text"] == "The number you are trying to verify is blacklisted for verification"


@responses.activate
def test_start_verification_blacklisted_error_with_request_id(client, dummy_data):
    stub(responses.POST, "https://api.nexmo.com/verify/json",
        fixture_path="verify/blocked_with_request_id.json"
    )

    params = {"number": "447525856424", "brand": "MyApp"}
    response = client.verify.start_verification(params)

    assert isinstance(response, dict)
    assert request_user_agent() == dummy_data.user_agent
    assert "number=447525856424" in request_body()
    assert "brand=MyApp" in request_body()
    assert response["status"] == "7"
    assert response["request_id"] == "12345678"
    assert response["error_text"] == "The number you are trying to verify is blacklisted for verification"


@responses.activate
def test_start_verification_blacklisted_error_with_network_and_request_id(client, dummy_data):
    stub(responses.POST, "https://api.nexmo.com/verify/json",
        fixture_path="verify/blocked_with_network_and_request_id.json"
    )

    params = {"number": "447525856424", "brand": "MyApp"}
    response = client.verify.start_verification(params)

    assert isinstance(response, dict)
    assert request_user_agent() == dummy_data.user_agent
    assert "number=447525856424" in request_body()
    assert "brand=MyApp" in request_body()
    assert response["status"] == "7"
    assert response["network"] == "25503"
    assert response["request_id"] == "12345678"
    assert response["error_text"] == "The number you are trying to verify is blacklisted for verification"

@responses.activate
def test_start_psd2_verification_blacklisted_error_with_network(client, dummy_data):
    stub(responses.POST, "https://api.nexmo.com/verify/psd2/json",
        fixture_path="verify/blocked_with_network.json"
    )

    params = {"number": "447525856424", "brand": "MyApp"}
    response = client.verify.psd2(params)

    assert isinstance(response, dict)
    assert request_user_agent() == dummy_data.user_agent
    assert "number=447525856424" in request_body()
    assert "brand=MyApp" in request_body()
    assert response["status"] == "7"
    assert response["network"] == "25503"
    assert response["error_text"] == "The number you are trying to verify is blacklisted for verification"


@responses.activate
def test_start_psd2_verification_blacklisted_error_with_request_id(client, dummy_data):
    stub(responses.POST, "https://api.nexmo.com/verify/psd2/json",
        fixture_path="verify/blocked_with_request_id.json"
    )

    params = {"number": "447525856424", "brand": "MyApp"}
    response = client.verify.psd2(params)

    assert isinstance(response, dict)
    assert request_user_agent() == dummy_data.user_agent
    assert "number=447525856424" in request_body()
    assert "brand=MyApp" in request_body()
    assert response["status"] == "7"
    assert response["request_id"] == "12345678"
    assert response["error_text"] == "The number you are trying to verify is blacklisted for verification"


@responses.activate
def test_start_psd2_verification_blacklisted_error_with_network_and_request_id(client, dummy_data):
    stub(responses.POST, "https://api.nexmo.com/verify/psd2/json",
        fixture_path="verify/blocked_with_network_and_request_id.json"
    )

    params = {"number": "447525856424", "brand": "MyApp"}
    response = client.verify.psd2(params)

    assert isinstance(response, dict)
    assert request_user_agent() == dummy_data.user_agent
    assert "number=447525856424" in request_body()
    assert "brand=MyApp" in request_body()
    assert response["status"] == "7"
    assert response["network"] == "25503"
    assert response["request_id"] == "12345678"
    assert response["error_text"] == "The number you are trying to verify is blacklisted for verification"



from vonage.errors import ProactiveConnectError, ClientError
from util import *

import responses
from pytest import raises
import csv


@responses.activate
def test_list_all_lists(proc, dummy_data):
    stub(
        responses.GET,
        'https://api-eu.vonage.com/v0.1/bulk/lists',
        fixture_path='proactive_connect/list_lists.json',
    )

    lists = proc.list_all_lists()
    assert request_user_agent() == dummy_data.user_agent
    assert lists['total_items'] == 2
    assert lists['_embedded']['lists'][0]['name'] == 'Recipients for demo'
    assert lists['_embedded']['lists'][0]['id'] == 'af8a84b6-c712-4252-ac8d-6e28ac9317ce'
    assert lists['_embedded']['lists'][1]['name'] == 'Salesforce contacts'
    assert lists['_embedded']['lists'][1]['datasource']['type'] == 'salesforce'


@responses.activate
def test_list_all_lists_options(proc):
    stub(
        responses.GET,
        'https://api-eu.vonage.com/v0.1/bulk/lists',
        fixture_path='proactive_connect/list_lists.json',
    )

    lists = proc.list_all_lists(page=1, page_size=5)
    assert lists['total_items'] == 2
    assert lists['_embedded']['lists'][0]['name'] == 'Recipients for demo'
    assert lists['_embedded']['lists'][0]['id'] == 'af8a84b6-c712-4252-ac8d-6e28ac9317ce'
    assert lists['_embedded']['lists'][1]['name'] == 'Salesforce contacts'


def test_pagination_errors(proc):
    with raises(ProactiveConnectError) as err:
        proc.list_all_lists(page=-1)
    assert str(err.value) == '"page" must be an int > 0.'

    with raises(ProactiveConnectError) as err:
        proc.list_all_lists(page_size=-1)
    assert str(err.value) == '"page_size" must be an int > 0.'


@responses.activate
def test_create_list_basic(proc):
    stub(
        responses.POST,
        'https://api-eu.vonage.com/v0.1/bulk/lists',
        fixture_path='proactive_connect/create_list_basic.json',
        status_code=201,
    )

    list = proc.create_list({'name': 'my_list'})
    assert list['id'] == '6994fd17-7691-4463-be16-172ab1430d97'
    assert list['name'] == 'my_list'


@responses.activate
def test_create_list_manual(proc):
    stub(
        responses.POST,
        'https://api-eu.vonage.com/v0.1/bulk/lists',
        fixture_path='proactive_connect/create_list_manual.json',
        status_code=201,
    )

    params = {
        "name": "my name",
        "description": "my description",
        "tags": ["vip", "sport"],
        "attributes": [{"name": "phone_number", "alias": "phone"}],
        "datasource": {"type": "manual"},
    }

    list = proc.create_list(params)
    assert list['id'] == '9508e7b8-fe99-4fdf-b022-65d7e461db2d'
    assert list['name'] == 'my_list'
    assert list['description'] == 'my description'
    assert list['tags'] == ['vip', 'sport']
    assert list['attributes'][0]['name'] == 'phone_number'


@responses.activate
def test_create_list_salesforce(proc):
    stub(
        responses.POST,
        'https://api-eu.vonage.com/v0.1/bulk/lists',
        fixture_path='proactive_connect/create_list_salesforce.json',
        status_code=201,
    )

    params = {
        "name": "my name",
        "description": "my description",
        "tags": ["vip", "sport"],
        "attributes": [{"name": "phone_number", "alias": "phone"}],
        "datasource": {
            "type": "salesforce",
            "integration_id": "salesforce_credentials",
            "soql": "select Id, LastName, FirstName, Phone, Email FROM Contact",
        },
    }

    list = proc.create_list(params)
    assert list['id'] == '246d17c4-79e6-4a25-8b4e-b777a83f6c30'
    assert list['name'] == 'my_salesforce_list'
    assert list['description'] == 'my salesforce description'
    assert list['datasource']['type'] == 'salesforce'
    assert list['datasource']['integration_id'] == 'salesforce_credentials'
    assert list['datasource']['soql'] == 'select Id, LastName, FirstName, Phone, Email FROM Contact'


def test_create_list_errors(proc):
    params = {
        "name": "my name",
        "datasource": {
            "type": "salesforce",
            "integration_id": 1234,
            "soql": "select Id, LastName, FirstName, Phone, Email FROM Contact",
        },
    }

    with raises(ProactiveConnectError) as err:
        proc.create_list({})
    assert str(err.value) == 'You must supply a name for the new list.'

    with raises(ProactiveConnectError) as err:
        proc.create_list(params)
    assert str(err.value) == 'You must supply values for "integration_id" and "soql" as strings.'

    with raises(ProactiveConnectError) as err:
        params['datasource'].pop('integration_id')
        proc.create_list(params)
    assert (
        str(err.value)
        == 'You must supply a value for "integration_id" and "soql" when creating a list with Salesforce.'
    )


@responses.activate
def test_create_list_invalid_name_error(proc):
    stub(
        responses.POST,
        'https://api-eu.vonage.com/v0.1/bulk/lists',
        fixture_path='proactive_connect/create_list_400.json',
        status_code=400,
    )

    with raises(ClientError) as err:
        proc.create_list({'name': 1234})
    assert (
        str(err.value)
        == 'Request data did not validate: Bad Request (https://developer.vonage.com/en/api-errors)\nError: name must be longer than or equal to 1 and shorter than or equal to 255 characters\nError: name must be a string'
    )


@responses.activate
def test_get_list(proc):
    list_id = '9508e7b8-fe99-4fdf-b022-65d7e461db2d'
    stub(
        responses.GET,
        f'https://api-eu.vonage.com/v0.1/bulk/lists/{list_id}',
        fixture_path='proactive_connect/get_list.json',
    )

    list = proc.get_list(list_id)
    assert list['id'] == '9508e7b8-fe99-4fdf-b022-65d7e461db2d'
    assert list['name'] == 'my_list'
    assert list['tags'] == ['vip', 'sport']


@responses.activate
def test_get_list_404(proc):
    list_id = 'a508e7b8-fe99-4fdf-b022-65d7e461db2d'
    stub(
        responses.GET,
        f'https://api-eu.vonage.com/v0.1/bulk/lists/{list_id}',
        fixture_path='proactive_connect/not_found.json',
        status_code=404,
    )

    with raises(ClientError) as err:
        proc.get_list(list_id)
    assert (
        str(err.value)
        == 'The requested resource does not exist: Not Found (https://developer.vonage.com/en/api-errors)'
    )


@responses.activate
def test_update_list(proc):
    list_id = '9508e7b8-fe99-4fdf-b022-65d7e461db2d'
    stub(
        responses.PUT,
        f'https://api-eu.vonage.com/v0.1/bulk/lists/{list_id}',
        fixture_path='proactive_connect/update_list.json',
    )

    params = {'name': 'my_list', 'tags': ['vip', 'sport', 'football']}
    list = proc.update_list(list_id, params)
    assert list['id'] == '9508e7b8-fe99-4fdf-b022-65d7e461db2d'
    assert list['tags'] == ['vip', 'sport', 'football']
    assert list['description'] == 'my updated description'
    assert list['updated_at'] == '2023-04-28T21:39:17.825Z'


@responses.activate
def test_update_list_salesforce(proc):
    list_id = '246d17c4-79e6-4a25-8b4e-b777a83f6c30'
    stub(
        responses.PUT,
        f'https://api-eu.vonage.com/v0.1/bulk/lists/{list_id}',
        fixture_path='proactive_connect/update_list_salesforce.json',
    )

    params = {'name': 'my_list', 'tags': ['music']}
    list = proc.update_list(list_id, params)
    assert list['id'] == list_id
    assert list['tags'] == ['music']
    assert list['updated_at'] == '2023-04-28T22:23:37.054Z'


def test_update_list_name_error(proc):
    with raises(ProactiveConnectError) as err:
        proc.update_list(
            '9508e7b8-fe99-4fdf-b022-65d7e461db2d', {'description': 'my new description'}
        )
    assert str(err.value) == 'You must supply a name for the new list.'


@responses.activate
def test_delete_list(proc):
    list_id = '9508e7b8-fe99-4fdf-b022-65d7e461db2d'
    stub(
        responses.DELETE,
        f'https://api-eu.vonage.com/v0.1/bulk/lists/{list_id}',
        fixture_path='null.json',
        status_code=204,
    )

    assert proc.delete_list(list_id) == None


@responses.activate
def test_delete_list_404(proc):
    list_id = '9508e7b8-fe99-4fdf-b022-65d7e461db2d'
    stub(
        responses.DELETE,
        f'https://api-eu.vonage.com/v0.1/bulk/lists/{list_id}',
        fixture_path='proactive_connect/not_found.json',
        status_code=404,
    )
    with raises(ClientError) as err:
        proc.delete_list(list_id)
    assert (
        str(err.value)
        == 'The requested resource does not exist: Not Found (https://developer.vonage.com/en/api-errors)'
    )


@responses.activate
def test_clear_list(proc):
    list_id = '9508e7b8-fe99-4fdf-b022-65d7e461db2d'
    stub(
        responses.POST,
        f'https://api-eu.vonage.com/v0.1/bulk/lists/{list_id}/clear',
        fixture_path='null.json',
        status_code=202,
    )

    assert proc.clear_list(list_id) == None


@responses.activate
def test_clear_list_404(proc):
    list_id = '9508e7b8-fe99-4fdf-b022-65d7e461db2d'
    stub(
        responses.POST,
        f'https://api-eu.vonage.com/v0.1/bulk/lists/{list_id}/clear',
        fixture_path='proactive_connect/not_found.json',
        status_code=404,
    )
    with raises(ClientError) as err:
        proc.clear_list(list_id)
    assert (
        str(err.value)
        == 'The requested resource does not exist: Not Found (https://developer.vonage.com/en/api-errors)'
    )


@responses.activate
def test_sync_list_from_datasource(proc):
    list_id = '246d17c4-79e6-4a25-8b4e-b777a83f6c30'
    stub(
        responses.POST,
        f'https://api-eu.vonage.com/v0.1/bulk/lists/{list_id}/fetch',
        fixture_path='null.json',
        status_code=202,
    )

    assert proc.sync_list_from_datasource(list_id) == None


@responses.activate
def test_sync_list_manual_datasource_error(proc):
    list_id = '9508e7b8-fe99-4fdf-b022-65d7e461db2d'
    stub(
        responses.POST,
        f'https://api-eu.vonage.com/v0.1/bulk/lists/{list_id}/fetch',
        fixture_path='proactive_connect/fetch_list_400.json',
        status_code=400,
    )

    with raises(ClientError) as err:
        proc.sync_list_from_datasource(list_id) == None
    assert (
        str(err.value)
        == 'Request data did not validate: Cannot Fetch a manual list (https://developer.vonage.com/en/api-errors)'
    )


@responses.activate
def test_sync_list_from_datasource_404(proc):
    list_id = '346d17c4-79e6-4a25-8b4e-b777a83f6c30'
    stub(
        responses.POST,
        f'https://api-eu.vonage.com/v0.1/bulk/lists/{list_id}/clear',
        fixture_path='proactive_connect/not_found.json',
        status_code=404,
    )
    with raises(ClientError) as err:
        proc.clear_list(list_id)
    assert (
        str(err.value)
        == 'The requested resource does not exist: Not Found (https://developer.vonage.com/en/api-errors)'
    )


@responses.activate
def test_list_all_items(proc):
    list_id = '9508e7b8-fe99-4fdf-b022-65d7e461db2d'
    stub(
        responses.GET,
        f'https://api-eu.vonage.com/v0.1/bulk/lists/{list_id}/items',
        fixture_path='proactive_connect/list_all_items.json',
    )

    items = proc.list_all_items(list_id, page=1, page_size=10)
    assert items['total_items'] == 2
    assert items['_embedded']['items'][0]['id'] == '04c7498c-bae9-40f9-bdcb-c4eabb0418fe'
    assert items['_embedded']['items'][1]['id'] == 'd91c39ed-7c34-4803-a139-34bb4b7c6d53'


@responses.activate
def test_list_all_items_error_not_found(proc):
    list_id = '9508e7b8-fe99-4fdf-b022-65d7e461db2d'
    stub(
        responses.GET,
        f'https://api-eu.vonage.com/v0.1/bulk/lists/{list_id}/items',
        fixture_path='proactive_connect/not_found.json',
        status_code=404,
    )

    with raises(ClientError) as err:
        proc.list_all_items(list_id)
    assert (
        str(err.value)
        == 'The requested resource does not exist: Not Found (https://developer.vonage.com/en/api-errors)'
    )


@responses.activate
def test_create_item(proc):
    list_id = '246d17c4-79e6-4a25-8b4e-b777a83f6c30'
    stub(
        responses.POST,
        f'https://api-eu.vonage.com/v0.1/bulk/lists/{list_id}/items',
        fixture_path='proactive_connect/item.json',
        status_code=201,
    )

    data = {'firstName': 'John', 'lastName': 'Doe', 'phone': '123456789101'}
    item = proc.create_item(list_id, data)

    assert item['id'] == 'd91c39ed-7c34-4803-a139-34bb4b7c6d53'
    assert item['data']['phone'] == '123456789101'


@responses.activate
def test_create_item_error_invalid_data(proc):
    list_id = '246d17c4-79e6-4a25-8b4e-b777a83f6c30'

    stub(
        responses.POST,
        f'https://api-eu.vonage.com/v0.1/bulk/lists/{list_id}/items',
        fixture_path='proactive_connect/item_400.json',
        status_code=400,
    )

    with raises(ClientError) as err:
        proc.create_item(list_id, {'data': 1234})
    assert (
        str(err.value)
        == 'Request data did not validate: Bad Request (https://developer.vonage.com/en/api-errors)\nError: data must be an object'
    )


@responses.activate
def test_create_item_error_not_found(proc):
    list_id = '346d17c4-79e6-4a25-8b4e-b777a83f6c30'
    stub(
        responses.POST,
        f'https://api-eu.vonage.com/v0.1/bulk/lists/{list_id}/items',
        fixture_path='proactive_connect/not_found.json',
        status_code=404,
    )

    data = {'firstName': 'John', 'lastName': 'Doe', 'phone': '123456789101'}
    with raises(ClientError) as err:
        proc.create_item(list_id, data)
    assert (
        str(err.value)
        == 'The requested resource does not exist: Not Found (https://developer.vonage.com/en/api-errors)'
    )


@responses.activate
def test_download_list_items(proc):
    list_id = '246d17c4-79e6-4a25-8b4e-b777a83f6c30'
    stub(
        responses.GET,
        f'https://api-eu.vonage.com/v0.1/bulk/lists/{list_id}/items/download',
        fixture_path='proactive_connect/list_items.csv',
    )

    proc.download_list_items(
        list_id, os.path.join(os.path.dirname(__file__), 'data/proactive_connect/list_items.csv')
    )
    items = _read_csv_file(
        os.path.join(os.path.dirname(__file__), 'data/proactive_connect/list_items.csv')
    )
    assert items[0]['favourite_number'] == '0'
    assert items[1]['least_favourite_number'] == '0'


@responses.activate
def test_download_list_items_error_not_found(proc):
    list_id = '346d17c4-79e6-4a25-8b4e-b777a83f6c30'
    stub(
        responses.GET,
        f'https://api-eu.vonage.com/v0.1/bulk/lists/{list_id}/items/download',
        fixture_path='proactive_connect/not_found.json',
        status_code=404,
    )

    with raises(ClientError) as err:
        proc.download_list_items(list_id, 'data/proactive_connect_list_items.csv')
    assert (
        str(err.value)
        == 'The requested resource does not exist: Not Found (https://developer.vonage.com/en/api-errors)'
    )


@responses.activate
def test_get_item(proc):
    list_id = '246d17c4-79e6-4a25-8b4e-b777a83f6c30'
    item_id = 'd91c39ed-7c34-4803-a139-34bb4b7c6d53'
    stub(
        responses.GET,
        f'https://api-eu.vonage.com/v0.1/bulk/lists/{list_id}/items/{item_id}',
        fixture_path='proactive_connect/item.json',
    )

    item = proc.get_item(list_id, item_id)
    assert item['id'] == 'd91c39ed-7c34-4803-a139-34bb4b7c6d53'
    assert item['data']['phone'] == '123456789101'


@responses.activate
def test_get_item_404(proc):
    list_id = '346d17c4-79e6-4a25-8b4e-b777a83f6c30'
    item_id = 'd91c39ed-7c34-4803-a139-34bb4b7c6d53'
    stub(
        responses.GET,
        f'https://api-eu.vonage.com/v0.1/bulk/lists/{list_id}/items/{item_id}',
        fixture_path='proactive_connect/not_found.json',
        status_code=404,
    )

    with raises(ClientError) as err:
        proc.get_item(list_id, item_id)
    assert (
        str(err.value)
        == 'The requested resource does not exist: Not Found (https://developer.vonage.com/en/api-errors)'
    )


@responses.activate
def test_update_item(proc):
    list_id = '246d17c4-79e6-4a25-8b4e-b777a83f6c30'
    item_id = 'd91c39ed-7c34-4803-a139-34bb4b7c6d53'
    data = {'first_name': 'John', 'last_name': 'Doe', 'phone': '447007000000'}
    stub(
        responses.PUT,
        f'https://api-eu.vonage.com/v0.1/bulk/lists/{list_id}/items/{item_id}',
        fixture_path='proactive_connect/update_item.json',
    )

    updated_item = proc.update_item(list_id, item_id, data)

    assert updated_item['id'] == item_id
    assert updated_item['data'] == data
    assert updated_item['updated_at'] == '2023-05-03T19:50:33.207Z'


@responses.activate
def test_update_item_error_invalid_data(proc):
    list_id = '246d17c4-79e6-4a25-8b4e-b777a83f6c30'
    item_id = 'd91c39ed-7c34-4803-a139-34bb4b7c6d53'
    data = 'asdf'
    stub(
        responses.PUT,
        f'https://api-eu.vonage.com/v0.1/bulk/lists/{list_id}/items/{item_id}',
        fixture_path='proactive_connect/item_400.json',
        status_code=400,
    )

    with raises(ClientError) as err:
        proc.update_item(list_id, item_id, data)
    assert (
        str(err.value)
        == 'Request data did not validate: Bad Request (https://developer.vonage.com/en/api-errors)\nError: data must be an object'
    )


@responses.activate
def test_update_item_404(proc):
    list_id = '346d17c4-79e6-4a25-8b4e-b777a83f6c30'
    item_id = 'd91c39ed-7c34-4803-a139-34bb4b7c6d53'
    data = {'first_name': 'John', 'last_name': 'Doe', 'phone': '447007000000'}
    stub(
        responses.PUT,
        f'https://api-eu.vonage.com/v0.1/bulk/lists/{list_id}/items/{item_id}',
        fixture_path='proactive_connect/not_found.json',
        status_code=404,
    )

    with raises(ClientError) as err:
        proc.update_item(list_id, item_id, data)
    assert (
        str(err.value)
        == 'The requested resource does not exist: Not Found (https://developer.vonage.com/en/api-errors)'
    )


@responses.activate
def test_delete_item(proc):
    list_id = '246d17c4-79e6-4a25-8b4e-b777a83f6c30'
    item_id = 'd91c39ed-7c34-4803-a139-34bb4b7c6d53'
    stub(
        responses.DELETE,
        f'https://api-eu.vonage.com/v0.1/bulk/lists/{list_id}/items/{item_id}',
        fixture_path='null.json',
        status_code=204,
    )

    response = proc.delete_item(list_id, item_id)
    assert response is None


@responses.activate
def test_delete_item_404(proc):
    list_id = '246d17c4-79e6-4a25-8b4e-b777a83f6c30'
    item_id = 'e91c39ed-7c34-4803-a139-34bb4b7c6d53'
    stub(
        responses.DELETE,
        f'https://api-eu.vonage.com/v0.1/bulk/lists/{list_id}/items/{item_id}',
        fixture_path='proactive_connect/not_found.json',
        status_code=404,
    )

    with raises(ClientError) as err:
        proc.delete_item(list_id, item_id)
    assert (
        str(err.value)
        == 'The requested resource does not exist: Not Found (https://developer.vonage.com/en/api-errors)'
    )


@responses.activate
def test_upload_list_items_from_csv(proc):
    list_id = '246d17c4-79e6-4a25-8b4e-b777a83f6c30'
    file_path = os.path.join(os.path.dirname(__file__), 'data/proactive_connect/csv_to_upload.csv')
    stub(
        responses.POST,
        f'https://api-eu.vonage.com/v0.1/bulk/lists/{list_id}/items/import',
        fixture_path='proactive_connect/upload_from_csv.json',
    )

    response = proc.upload_list_items(list_id, file_path)
    assert response['inserted'] == 3


@responses.activate
def test_upload_list_items_from_csv_404(proc):
    list_id = '346d17c4-79e6-4a25-8b4e-b777a83f6c30'
    file_path = os.path.join(os.path.dirname(__file__), 'data/proactive_connect/csv_to_upload.csv')
    stub(
        responses.POST,
        f'https://api-eu.vonage.com/v0.1/bulk/lists/{list_id}/items/import',
        fixture_path='proactive_connect/not_found.json',
        status_code=404,
    )

    with raises(ClientError) as err:
        proc.upload_list_items(list_id, file_path)
    assert (
        str(err.value)
        == 'The requested resource does not exist: Not Found (https://developer.vonage.com/en/api-errors)'
    )


@responses.activate
def test_list_events(proc):
    stub(
        responses.GET,
        'https://api-eu.vonage.com/v0.1/bulk/events',
        fixture_path='proactive_connect/list_events.json',
    )

    lists = proc.list_events()
    assert lists['total_items'] == 1
    assert lists['_embedded']['events'][0]['occurred_at'] == '2022-08-07T13:18:21.970Z'
    assert lists['_embedded']['events'][0]['type'] == 'action-call-succeeded'
    assert lists['_embedded']['events'][0]['run_id'] == '7d0d4e5f-6453-4c63-87cf-f95b04377324'


def _read_csv_file(path):
    with open(os.path.join(os.path.dirname(__file__), path)) as csv_file:
        reader = csv.DictReader(csv_file)
        dict_list = [row for row in reader]
        return dict_list



from .errors import PricingTypeError

from deprecated import deprecated
class Account:
    account_auth_type = 'params'
    pricing_auth_type = 'params'
    secrets_auth_type = 'header'

    allowed_pricing_types = {'sms', 'sms-transit', 'voice'}

    def __init__(self, client):
        self._client = client

    def get_balance(self):
        return self._client.get(self._client.host(), "/account/get-balance", auth_type=Account.account_auth_type)

    def topup(self, params=None, **kwargs):
        return self._client.post(
            self._client.host(), 
            "/account/top-up",
            params or kwargs, 
            auth_type=Account.account_auth_type,
            body_is_json=False,
        )

    def get_country_pricing(self, country_code: str, type: str = 'sms'):
        self._check_allowed_pricing_type(type)
        return self._client.get(
            self._client.host(), 
            f"/account/get-pricing/outbound/{type}", 
            {"country": country_code},
            auth_type=Account.pricing_auth_type
        )

    def get_all_countries_pricing(self, type: str = 'sms'):
        self._check_allowed_pricing_type(type)
        return self._client.get(
            self._client.host(), f"/account/get-full-pricing/outbound/{type}", auth_type=Account.pricing_auth_type
        )

    def get_prefix_pricing(self, prefix: str, type: str = 'sms'):
        self._check_allowed_pricing_type(type)
        return self._client.get(
            self._client.host(), 
            f"/account/get-prefix-pricing/outbound/{type}",
            {"prefix": prefix},
            auth_type=Account.pricing_auth_type,
        )

    @deprecated(version='3.0.0', reason='The "account/get-phone-pricing" endpoint is deprecated.')
    def get_sms_pricing(self, number: str):
        return self._client.get(
            self._client.host(), 
            "/account/get-phone-pricing/outbound/sms", 
            {"phone": number},
            auth_type=Account.pricing_auth_type,
        )

    @deprecated(version='3.0.0', reason='The "account/get-phone-pricing" endpoint is deprecated.')
    def get_voice_pricing(self, number: str):
        return self._client.get(
            self._client.host(), 
            "/account/get-phone-pricing/outbound/voice", 
            {"phone": number},
            auth_type=Account.pricing_auth_type,
        )

    def update_default_sms_webhook(self, params=None, **kwargs):
        return self._client.post(
            self._client.host(), 
            "/account/settings", 
            params or kwargs,
            auth_type=Account.account_auth_type,
            body_is_json=False,
        )

    def list_secrets(self, api_key):
        return self._client.get(
            self._client.api_host(),
            f"/accounts/{api_key}/secrets",
            auth_type=Account.secrets_auth_type,
        )

    def get_secret(self, api_key, secret_id):
        return self._client.get(
            self._client.api_host(),
            f"/accounts/{api_key}/secrets/{secret_id}",
            auth_type=Account.secrets_auth_type,
        )

    def create_secret(self, api_key, secret):
        body = {"secret": secret}
        return self._client.post(
            self._client.api_host(), 
            f"/accounts/{api_key}/secrets",
            body,
            auth_type=Account.secrets_auth_type,
            body_is_json=False,
        )

    def revoke_secret(self, api_key, secret_id):
        return self._client.delete(
            self._client.api_host(),
            f"/accounts/{api_key}/secrets/{secret_id}",
            auth_type=Account.secrets_auth_type,
        )

    def _check_allowed_pricing_type(self, type):
        if type not in Account.allowed_pricing_types:
            raise PricingTypeError('Invalid pricing type specified.')



from pydantic import BaseModel, validator
from typing import Optional, Dict
from typing_extensions import Literal


class PayPrompts:
    class VoicePrompt(BaseModel):
        language: Optional[str]
        style: Optional[int]

    class TextPrompt(BaseModel):
        type: Literal['CardNumber', 'ExpirationDate', 'SecurityCode']
        text: str
        errors: Dict[
            Literal['InvalidCardType', 'InvalidCardNumber', 'InvalidExpirationDate', 'InvalidSecurityCode', 'Timeout'],
            Dict[Literal['text'], str],
        ]

        @validator('errors')
        def check_valid_error_format(cls, v, values):
            if values['type'] == 'CardNumber':
                allowed_values = {'InvalidCardType', 'InvalidCardNumber', 'Timeout'}
                cls.check_allowed_values(v, allowed_values, values['type'])
            elif values['type'] == 'ExpirationDate':
                allowed_values = {'InvalidExpirationDate', 'Timeout'}
                cls.check_allowed_values(v, allowed_values, values['type'])
            elif values['type'] == 'SecurityCode':
                allowed_values = {'InvalidSecurityCode', 'Timeout'}
                cls.check_allowed_values(v, allowed_values, values['type'])
            return v

        def check_allowed_values(errors, allowed_values, prompt_type):
            for key in errors:
                if key not in allowed_values:
                    raise ValueError(f'Value "{key}" is not a valid error for the "{prompt_type}" prompt type.')

    @classmethod
    def create_voice_model(cls, dict) -> VoicePrompt:
        return cls.VoicePrompt.parse_obj(dict)

    @classmethod
    def create_text_model(cls, dict) -> TextPrompt:
        return cls.TextPrompt.parse_obj(dict)



from vonage import Client, Verify2
from util import *
from vonage.errors import ClientError, Verify2Error

from pydantic import ValidationError
from pytest import raises
import responses

verify2 = Verify2(Client())


@responses.activate
def test_new_request_sms_basic(dummy_data):
    stub(
        responses.POST,
        'https://api.nexmo.com/v2/verify',
        fixture_path='verify2/create_request.json',
        status_code=202,
    )

    params = {'brand': 'ACME, Inc', 'workflow': [{'channel': 'sms', 'to': '447700900000'}]}
    verify_request = verify2.new_request(params)

    assert request_user_agent() == dummy_data.user_agent
    assert verify_request['request_id'] == 'c11236f4-00bf-4b89-84ba-88b25df97315'


@responses.activate
def test_new_request_sms_full():
    stub(
        responses.POST,
        'https://api.nexmo.com/v2/verify',
        fixture_path='verify2/create_request.json',
        status_code=202,
    )

    params = {
        'locale': 'en-gb',
        'channel_timeout': 120,
        'client_ref': 'my client ref',
        'code_length': 8,
        'fraud_check': False,
        'brand': 'ACME, Inc',
        'workflow': [{'channel': 'sms', 'to': '447700900000', 'app_hash': 'asdfghjklqw'}],
    }
    verify_request = verify2.new_request(params)

    assert verify_request['request_id'] == 'c11236f4-00bf-4b89-84ba-88b25df97315'


@responses.activate
def test_new_request_sms_custom_code(dummy_data):
    stub(
        responses.POST,
        'https://api.nexmo.com/v2/verify',
        fixture_path='verify2/create_request.json',
        status_code=202,
    )

    params = {
        'brand': 'ACME, Inc',
        'code': 'asdfghjk',
        'workflow': [{'channel': 'sms', 'to': '447700900000'}],
    }
    verify_request = verify2.new_request(params)

    assert request_user_agent() == dummy_data.user_agent
    assert verify_request['request_id'] == 'c11236f4-00bf-4b89-84ba-88b25df97315'


@responses.activate
def test_new_request_error_fraud_check_invalid_account(dummy_data):
    stub(
        responses.POST,
        'https://api.nexmo.com/v2/verify',
        fixture_path='verify2/fraud_check_invalid_account.json',
        status_code=403,
    )

    params = {
        'brand': 'ACME, Inc',
        'fraud_check': False,
        'workflow': [{'channel': 'sms', 'to': '447700900000'}],
    }

    with raises(ClientError) as err:
        verify2.new_request(params)
    assert (
        str(err.value)
        == 'Forbidden: Your account does not have permission to perform this action. (https://developer.nexmo.com/api-errors#forbidden)'
    )


def test_new_request_sms_custom_code_length_error():
    params = {
        'code_length': 4,
        'brand': 'ACME, Inc',
        'code': 'a',
        'workflow': [{'channel': 'sms', 'to': '447700900000'}],
    }

    with raises(ValidationError) as err:
        verify2.new_request(params)
    assert 'ensure this value has at least 4 characters' in str(err.value)


def test_new_request_sms_custom_code_character_error():
    params = {
        'code_length': 4,
        'brand': 'ACME, Inc',
        'code': '?!@%',
        'workflow': [{'channel': 'sms', 'to': '447700900000'}],
    }

    with raises(ValidationError) as err:
        verify2.new_request(params)
    assert 'string does not match regex' in str(err.value)


def test_new_request_invalid_channel_error():
    params = {
        'code_length': 4,
        'brand': 'ACME, Inc',
        'workflow': [{'channel': 'carrier_pigeon', 'to': '447700900000'}],
    }

    with raises(Verify2Error) as err:
        verify2.new_request(params)
    assert (
        str(err.value)
        == 'You must specify a valid verify channel inside the "workflow" object, one of: "[\'sms\', \'whatsapp\', \'whatsapp_interactive\', \'voice\', \'email\', \'silent_auth\']"'
    )


def test_new_request_code_length_error():
    params = {
        'code_length': 1000,
        'brand': 'ACME, Inc',
        'workflow': [{'channel': 'sms', 'to': '447700900000'}],
    }

    with raises(ValidationError) as err:
        verify2.new_request(params)
    assert 'ensure this value is less than or equal to 10' in str(err.value)


def test_new_request_to_error():
    params = {
        'brand': 'ACME, Inc',
        'workflow': [{'channel': 'sms', 'to': '123'}],
    }

    with raises(Verify2Error) as err:
        verify2.new_request(params)
    assert 'You must specify a valid "to" value for channel "sms"' in str(err.value)


def test_new_request_sms_app_hash_error():
    params = {
        'brand': 'ACME, Inc',
        'workflow': [{'channel': 'sms', 'to': '447700900000', 'app_hash': '00'}],
    }

    with raises(Verify2Error) as err:
        verify2.new_request(params)
    assert 'Invalid "app_hash" specified.' in str(err.value)


def test_new_request_whatsapp_app_hash_error():
    params = {
        'brand': 'ACME, Inc',
        'workflow': [{'channel': 'whatsapp', 'to': '447700900000', 'app_hash': 'asdfqwerzxc'}],
    }

    with raises(Verify2Error) as err:
        verify2.new_request(params)
    assert (
        str(err.value)
        == 'Cannot specify a value for "app_hash" unless using SMS for authentication.'
    )


@responses.activate
def test_new_request_whatsapp():
    stub(
        responses.POST,
        'https://api.nexmo.com/v2/verify',
        fixture_path='verify2/create_request.json',
        status_code=202,
    )

    params = {'brand': 'ACME, Inc', 'workflow': [{'channel': 'whatsapp', 'to': '447700900000'}]}
    verify_request = verify2.new_request(params)

    assert verify_request['request_id'] == 'c11236f4-00bf-4b89-84ba-88b25df97315'


@responses.activate
def test_new_request_whatsapp_custom_code():
    stub(
        responses.POST,
        'https://api.nexmo.com/v2/verify',
        fixture_path='verify2/create_request.json',
        status_code=202,
    )

    params = {
        'brand': 'ACME, Inc',
        'code': 'asdfghjk',
        'workflow': [{'channel': 'whatsapp', 'to': '447700900000'}],
    }
    verify_request = verify2.new_request(params)

    assert verify_request['request_id'] == 'c11236f4-00bf-4b89-84ba-88b25df97315'


@responses.activate
def test_new_request_whatsapp_from_field():
    stub(
        responses.POST,
        'https://api.nexmo.com/v2/verify',
        fixture_path='verify2/create_request.json',
        status_code=202,
    )

    params = {
        'brand': 'ACME, Inc',
        'workflow': [{'channel': 'whatsapp', 'to': '447700900000', 'from': '447000000000'}],
    }
    verify_request = verify2.new_request(params)

    assert verify_request['request_id'] == 'c11236f4-00bf-4b89-84ba-88b25df97315'


@responses.activate
def test_new_request_whatsapp_invalid_sender_error():
    stub(
        responses.POST,
        'https://api.nexmo.com/v2/verify',
        fixture_path='verify2/invalid_sender.json',
        status_code=422,
    )

    params = {
        'brand': 'ACME, Inc',
        'workflow': [{'channel': 'whatsapp', 'to': '447700900000', 'from': 'asdfghjkl'}],
    }
    with pytest.raises(ClientError) as err:
        verify2.new_request(params)
    assert str(err.value) == 'You must specify a valid "from" value if included.'


@responses.activate
def test_new_request_whatsapp_sender_unregistered_error():
    stub(
        responses.POST,
        'https://api.nexmo.com/v2/verify',
        fixture_path='verify2/invalid_sender.json',
        status_code=422,
    )

    params = {
        'brand': 'ACME, Inc',
        'workflow': [{'channel': 'whatsapp', 'to': '447700900000', 'from': '447999999999'}],
    }
    with pytest.raises(ClientError) as err:
        verify2.new_request(params)
    assert (
        str(err.value)
        == 'Invalid sender: The `from` parameter is invalid. (https://developer.nexmo.com/api-errors#invalid-param)'
    )


@responses.activate
def test_new_request_whatsapp_interactive():
    stub(
        responses.POST,
        'https://api.nexmo.com/v2/verify',
        fixture_path='verify2/create_request.json',
        status_code=202,
    )

    params = {
        'brand': 'ACME, Inc',
        'workflow': [{'channel': 'whatsapp_interactive', 'to': '447700900000'}],
    }
    verify_request = verify2.new_request(params)

    assert verify_request['request_id'] == 'c11236f4-00bf-4b89-84ba-88b25df97315'


@responses.activate
def test_new_request_voice():
    stub(
        responses.POST,
        'https://api.nexmo.com/v2/verify',
        fixture_path='verify2/create_request.json',
        status_code=202,
    )

    params = {'brand': 'ACME, Inc', 'workflow': [{'channel': 'voice', 'to': '447700900000'}]}
    verify_request = verify2.new_request(params)

    assert verify_request['request_id'] == 'c11236f4-00bf-4b89-84ba-88b25df97315'


@responses.activate
def test_new_request_voice_custom_code():
    stub(
        responses.POST,
        'https://api.nexmo.com/v2/verify',
        fixture_path='verify2/create_request.json',
        status_code=202,
    )

    params = {
        'brand': 'ACME, Inc',
        'code': 'asdfhjkl',
        'workflow': [{'channel': 'voice', 'to': '447700900000'}],
    }
    verify_request = verify2.new_request(params)

    assert verify_request['request_id'] == 'c11236f4-00bf-4b89-84ba-88b25df97315'


@responses.activate
def test_new_request_email():
    stub(
        responses.POST,
        'https://api.nexmo.com/v2/verify',
        fixture_path='verify2/create_request.json',
        status_code=202,
    )

    params = {
        'brand': 'ACME, Inc',
        'workflow': [{'channel': 'email', 'to': 'recipient@example.com'}],
    }
    verify_request = verify2.new_request(params)

    assert verify_request['request_id'] == 'c11236f4-00bf-4b89-84ba-88b25df97315'


@responses.activate
def test_new_request_email_additional_fields():
    stub(
        responses.POST,
        'https://api.nexmo.com/v2/verify',
        fixture_path='verify2/create_request.json',
        status_code=202,
    )

    params = {
        'locale': 'en-gb',
        'channel_timeout': 120,
        'client_ref': 'my client ref',
        'code_length': 8,
        'brand': 'ACME, Inc',
        'code': 'asdfhjkl',
        'workflow': [
            {'channel': 'email', 'to': 'recipient@example.com', 'from': 'sender@example.com'}
        ],
    }
    verify_request = verify2.new_request(params)

    assert verify_request['request_id'] == 'c11236f4-00bf-4b89-84ba-88b25df97315'


@responses.activate
def test_new_request_email_error():
    stub(
        responses.POST,
        'https://api.nexmo.com/v2/verify',
        fixture_path='verify2/invalid_email.json',
        status_code=422,
    )

    params = {
        'brand': 'ACME, Inc',
        'workflow': [{'channel': 'email', 'to': 'not-an-email-address'}],
    }
    with pytest.raises(ClientError) as err:
        verify2.new_request(params)
    assert (
        str(err.value)
        == 'Invalid params: The value of one or more parameters is invalid (https://www.nexmo.com/messages/Errors#InvalidParams)'
    )


@responses.activate
def test_new_request_silent_auth():
    stub(
        responses.POST,
        'https://api.nexmo.com/v2/verify',
        fixture_path='verify2/create_request.json',
        status_code=202,
    )

    params = {'brand': 'ACME, Inc', 'workflow': [{'channel': 'silent_auth', 'to': '447700900000'}]}
    verify_request = verify2.new_request(params)

    assert verify_request['request_id'] == 'c11236f4-00bf-4b89-84ba-88b25df97315'


@responses.activate
def test_new_request_error_conflict():
    stub(
        responses.POST,
        'https://api.nexmo.com/v2/verify',
        fixture_path='verify2/error_conflict.json',
        status_code=409,
    )
    params = {'brand': 'ACME, Inc', 'workflow': [{'channel': 'sms', 'to': '447700900000'}]}

    with raises(ClientError) as err:
        verify2.new_request(params)
    assert (
        str(err.value)
        == "Conflict: Concurrent verifications to the same number are not allowed. (https://www.developer.vonage.com/api-errors/verify#conflict)"
    )


@responses.activate
def test_new_request_rate_limit():
    stub(
        responses.POST,
        'https://api.nexmo.com/v2/verify',
        fixture_path='verify2/rate_limit.json',
        status_code=429,
    )
    params = {'brand': 'ACME, Inc', 'workflow': [{'channel': 'sms', 'to': '447700900000'}]}

    with raises(ClientError) as err:
        verify2.new_request(params)
    assert (
        str(err.value)
        == "Rate Limit Hit: Please wait, then retry your request (https://www.developer.vonage.com/api-errors#throttled)"
    )


@responses.activate
def test_check_code():
    stub(
        responses.POST,
        'https://api.nexmo.com/v2/verify/c11236f4-00bf-4b89-84ba-88b25df97315',
        fixture_path='verify2/check_code.json',
    )

    response = verify2.check_code('c11236f4-00bf-4b89-84ba-88b25df97315', '1234')
    assert response['request_id'] == 'e043d872-459b-4750-a20c-d33f91d6959f'
    assert response['status'] == 'completed'


@responses.activate
def test_check_code_invalid_code():
    stub(
        responses.POST,
        'https://api.nexmo.com/v2/verify/c11236f4-00bf-4b89-84ba-88b25df97315',
        fixture_path='verify2/invalid_code.json',
        status_code=400,
    )

    with pytest.raises(ClientError) as err:
        verify2.check_code('c11236f4-00bf-4b89-84ba-88b25df97315', '5678')

    assert (
        str(err.value)
        == 'Invalid Code: The code you provided does not match the expected value. (https://developer.nexmo.com/api-errors#bad-request)'
    )


@responses.activate
def test_check_code_already_verified():
    stub(
        responses.POST,
        'https://api.nexmo.com/v2/verify/c11236f4-00bf-4b89-84ba-88b25df97315',
        fixture_path='verify2/already_verified.json',
        status_code=404,
    )

    with pytest.raises(ClientError) as err:
        verify2.check_code('c11236f4-00bf-4b89-84ba-88b25df97315', '5678')

    assert (
        str(err.value)
        == "Not Found: Request '5fcc26ef-1e54-48a6-83ab-c47546a19824' was not found or it has been verified already. (https://developer.nexmo.com/api-errors#not-found)"
    )


@responses.activate
def test_check_code_workflow_not_supported():
    stub(
        responses.POST,
        'https://api.nexmo.com/v2/verify/c11236f4-00bf-4b89-84ba-88b25df97315',
        fixture_path='verify2/code_not_supported.json',
        status_code=409,
    )

    with pytest.raises(ClientError) as err:
        verify2.check_code('c11236f4-00bf-4b89-84ba-88b25df97315', '5678')

    assert (
        str(err.value)
        == 'Conflict: The current Verify workflow step does not support a code. (https://developer.nexmo.com/api-errors#conflict)'
    )


@responses.activate
def test_check_code_too_many_invalid_code_attempts():
    stub(
        responses.POST,
        'https://api.nexmo.com/v2/verify/c11236f4-00bf-4b89-84ba-88b25df97315',
        fixture_path='verify2/too_many_code_attempts.json',
        status_code=410,
    )

    with pytest.raises(ClientError) as err:
        verify2.check_code('c11236f4-00bf-4b89-84ba-88b25df97315', '5678')

    assert (
        str(err.value)
        == 'Invalid Code: An incorrect code has been provided too many times. Workflow terminated. (https://developer.nexmo.com/api-errors#gone)'
    )


@responses.activate
def test_check_code_rate_limit():
    stub(
        responses.POST,
        'https://api.nexmo.com/v2/verify/c11236f4-00bf-4b89-84ba-88b25df97315',
        fixture_path='verify2/rate_limit.json',
        status_code=429,
    )

    with raises(ClientError) as err:
        verify2.check_code('c11236f4-00bf-4b89-84ba-88b25df97315', '5678')
    assert (
        str(err.value)
        == "Rate Limit Hit: Please wait, then retry your request (https://www.developer.vonage.com/api-errors#throttled)"
    )


@responses.activate
def test_cancel_verification():
    stub(
        responses.DELETE,
        'https://api.nexmo.com/v2/verify/c11236f4-00bf-4b89-84ba-88b25df97315',
        fixture_path='no_content.json',
        status_code=204,
    )

    assert verify2.cancel_verification('c11236f4-00bf-4b89-84ba-88b25df97315') == None


@responses.activate
def test_cancel_verification_error_not_found():
    stub(
        responses.DELETE,
        'https://api.nexmo.com/v2/verify/c11236f4-00bf-4b89-84ba-88b25df97315',
        fixture_path='verify2/request_not_found.json',
        status_code=404,
    )

    with raises(ClientError) as err:
        verify2.cancel_verification('c11236f4-00bf-4b89-84ba-88b25df97315')
    assert (
        str(err.value)
        == "Not Found: Request 'c11236f4-00bf-4b89-84ba-88b25df97315' was not found or it has been verified already. (https://developer.nexmo.com/api-errors#not-found)"
    )


@responses.activate
def test_new_request_multiple_workflows():
    stub(
        responses.POST,
        'https://api.nexmo.com/v2/verify',
        fixture_path='verify2/create_request.json',
        status_code=202,
    )

    params = {
        'brand': 'ACME, Inc',
        'workflow': [
            {'channel': 'whatsapp_interactive', 'to': '447700900000'},
            {'channel': 'sms', 'to': '4477009999999'},
        ],
    }
    verify_request = verify2.new_request(params)

    assert verify_request['request_id'] == 'c11236f4-00bf-4b89-84ba-88b25df97315'


def test_remove_unnecessary_fraud_check():
    params = {
        'brand': 'ACME, Inc',
        'workflow': [{'channel': 'sms', 'to': '447700900000'}],
        'fraud_check': True,
    }
    verify2._remove_unnecessary_fraud_check(params)

    assert 'fraud_check' not in params



import pytz
from datetime import datetime
from ._internal import _format_date_param

class Sms:
    defaults = {'auth_type': 'params', 'body_is_json': False}

    def __init__(self, client):
        self._client = client
    
    def send_message(self, params):
        """
        Send an SMS message.
        Requires a client initialized with `key` and either `secret` or `signature_secret`.
        :param dict params: A dict of values described at `Send an SMS <https://developer.vonage.com/api/sms#send-an-sms>`_
        """
        return self._client.post(
            self._client.host(), 
            "/sms/json", 
            params, 
            supports_signature_auth=True,
            **Sms.defaults,
        )
    
    def submit_sms_conversion(self, message_id, delivered=True, timestamp=None):
        """
        Notify Vonage that an SMS was successfully received.

        If you are using the Verify API for 2FA, this information is sent to Vonage automatically
        so you do not need to use this method to submit conversion data about 2FA messages.

        :param message_id: The `message-id` str returned by the send_message call.
        :param delivered: A `bool` indicating that the message was or was not successfully delivered.
        :param timestamp: A `datetime` object containing the time the SMS arrived.
        :return: The parsed response from the server. On success, the bytestring b'OK'
        """
        params = {
            "message-id": message_id,
            "delivered": delivered,
            "timestamp": timestamp or datetime.now(pytz.utc)
        }
        # Ensure timestamp is a string:
        _format_date_param(params, "timestamp")
        return self._client.post(self._client.api_host(), "/conversions/sms", params, **Sms.defaults)
